# noinspection PyPep8Naming
# This file takes some liberties to match the naming from the dyld implementation
from __future__ import annotations

import ctypes
import enum
from ctypes import Structure, c_uint16, c_uint32, c_uint64

# Some type aliases
FilePointer = int  # Offset into a raw binary file
FileOffset = int  # Offset to another FilePointer
MemoryPointer = int  # Offset into the mapped memory space

# All of this is based on the dyld-852.2 source code


class HelperStruct(Structure):
    """
    Subclass of ctypes.Structure that adds a helpful repr method for debugging
    """

    def __repr__(self):
        r = [f"<struct {self.__class__.__name__}"]
        for name, *_ in self._fields_:
            r.append(f"{name}={hex(getattr(self, name))}")
        r.append(">")
        return " ".join(r)


class DyldImportFormats(enum.IntEnum):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L249-L254
    """

    DYLD_CHAINED_IMPORT = (1,)
    DYLD_CHAINED_IMPORT_ADDEND = (2,)
    DYLD_CHAINED_IMPORT_ADDEND64 = (3,)


class DyldChainedPtrFormats(enum.IntEnum):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L89-L104

    """

    DYLD_CHAINED_PTR_ARM64E = 1  # stride 8, unauth target is vmaddr
    DYLD_CHAINED_PTR_64 = 2  # target is vmaddr
    DYLD_CHAINED_PTR_32 = 3
    DYLD_CHAINED_PTR_32_CACHE = 4
    DYLD_CHAINED_PTR_32_FIRMWARE = 5
    DYLD_CHAINED_PTR_64_OFFSET = 6  # target is vm offset
    # DYLD_CHAINED_PTR_ARM64E_OFFSET          =  7 # old name
    DYLD_CHAINED_PTR_ARM64E_KERNEL = 7  # stride 4, unauth target is vm offset
    DYLD_CHAINED_PTR_64_KERNEL_CACHE = 8
    DYLD_CHAINED_PTR_ARM64E_USERLAND = 9  # stride 8, unauth target is vm offset
    DYLD_CHAINED_PTR_ARM64E_FIRMWARE = 10  # stride 4, unauth target is vmaddr
    DYLD_CHAINED_PTR_X86_64_KERNEL_CACHE = 11  # stride 1, x86_64 kernel caches
    DYLD_CHAINED_PTR_ARM64E_USERLAND24 = 12  # stride 8, unauth target is vm offset, 24-bit bind


# noinspection PyPep8Naming
class dyld_chained_ptr_arm64e_auth_rebase(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L128-L138
    """

    _fields_ = [
        ("target", c_uint64, 32),
        ("diversity", c_uint64, 16),
        ("addrDiv", c_uint64, 1),
        ("key", c_uint64, 2),
        ("next", c_uint64, 11),
        ("bind", c_uint64, 1),
        ("auth", c_uint64, 1),
    ]


# noinspection PyPep8Naming
class dyld_chained_ptr_arm64e_auth_bind(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L140-L151
    """

    _fields_ = [
        ("ordinal", c_uint64, 16),
        ("zero", c_uint64, 16),
        ("diversity", c_uint64, 16),
        ("addrDiv", c_uint64, 1),
        ("key", c_uint64, 2),
        ("next", c_uint64, 11),
        ("bind", c_uint64, 1),
        ("auth", c_uint64, 1),
    ]


# noinspection PyPep8Naming
class dyld_chained_ptr_arm64e_rebase(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L107-L115
    """

    _fields_ = [
        ("target", c_uint64, 43),
        ("high8", c_uint64, 8),
        ("next", c_uint64, 11),
        ("bind", c_uint64, 1),
        ("auth", c_uint64, 1),
    ]


# noinspection PyPep8Naming
class dyld_chained_ptr_arm64e_bind(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L117-L126
    """

    _fields_ = [
        ("ordinal", c_uint64, 16),
        ("zero", c_uint64, 16),
        ("addend", c_uint64, 19),
        ("next", c_uint64, 11),
        ("bind", c_uint64, 1),
        ("auth", c_uint64, 1),
    ]


# noinspection PyPep8Naming
class dyld_chained_ptr_arm64e_bind24(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L164-L173
    """


# noinspection PyPep8Naming
class dyld_chained_ptr_arm64e_auth_bind24(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L175-L186
    """

    _fields_ = [
        ("ordinal", c_uint64, 24),
        ("zero", c_uint64, 8),
        ("diversity", c_uint64, 16),
        ("addrDiv", c_uint64, 1),
        ("key", c_uint64, 2),
        ("next", c_uint64, 11),
        ("bind", c_uint64, 1),
        ("auth", c_uint64, 1),
    ]


class Arm64e(ctypes.Union):
    """
    named after the Union `Arm64e` from dyld MachOLoaded.h
    https://github.com/apple-opensource/dyld/blob/852.2/dyld3/MachOLoaded.h#L89-L103
    """

    authRebase: dyld_chained_ptr_arm64e_auth_rebase
    authBind: dyld_chained_ptr_arm64e_auth_bind
    rebase: dyld_chained_ptr_arm64e_rebase
    bind: dyld_chained_ptr_arm64e_bind
    bind24: dyld_chained_ptr_arm64e_bind24
    authBind24: dyld_chained_ptr_arm64e_auth_bind24

    _fields_ = [
        ("authRebase", dyld_chained_ptr_arm64e_auth_rebase),
        ("authBind", dyld_chained_ptr_arm64e_auth_bind),
        ("rebase", dyld_chained_ptr_arm64e_rebase),
        ("bind", dyld_chained_ptr_arm64e_bind),
        ("bind24", dyld_chained_ptr_arm64e_bind24),
        ("authBind24", dyld_chained_ptr_arm64e_auth_bind24),
    ]

    @property
    def sign_extended_addend(self):
        assert self.authBind.bind == 1
        assert self.authBind.auth == 0
        addend19 = self.bind.addend
        if addend19 & 0x40000:
            return addend19 | 0xFFFFFFFFFFFC0000
        else:
            return addend19

    @property
    def unpack_target(self):
        assert self.authBind.bind == 0
        assert self.authBind.auth == 0
        return self.rebase.high8 << 56 | self.rebase.target

    @staticmethod
    def check_valid_pointer_format(pointer_format: DyldChainedPtrFormats) -> bool:
        """
        helper to check if a pointer format is relevant for this
        :param pointer_format:
        :return:
        """
        return pointer_format in [
            DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E,
            DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E_USERLAND,
            DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E_USERLAND24,
            DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E_KERNEL,
            DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E_FIRMWARE,
        ]


# noinspection PyPep8Naming
class dyld_chained_ptr_64_rebase(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L153-L161
    """

    target: FilePointer | FileOffset
    high8: int
    next: int
    bind: int

    _fields_ = [
        ("target", c_uint64, 36),
        ("high8", c_uint64, 8),
        ("_reserved", c_uint64, 7),
        ("next", c_uint64, 12),
        ("bind", c_uint64, 1),
    ]

    @property
    def unpackedTarget(self):
        return self.high8 << 56 | self.target


# noinspection PyPep8Naming
class dyld_chained_ptr_64_bind(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L189-L197
    """

    ordinal: int
    addend: int
    next: int
    bind: int

    _fields_ = [
        ("ordinal", c_uint64, 24),
        ("addend", c_uint64, 8),
        ("_reserved", c_uint64, 19),
        ("next", c_uint64, 12),
        ("bind", c_uint64, 1),
    ]


class Generic64(ctypes.Union):
    """
    named after the Union `Generic64` from dyld MachOLoaded.h
    https://github.com/apple-opensource/dyld/blob/852.2/dyld3/MachOLoaded.h#L105-L111
    """

    rebase: dyld_chained_ptr_64_rebase
    bind: dyld_chained_ptr_64_bind

    _fields_ = [
        ("rebase", dyld_chained_ptr_64_rebase),
        ("bind", dyld_chained_ptr_64_bind),
    ]

    @staticmethod
    def check_valid_pointer_format(pointer_format: DyldChainedPtrFormats) -> bool:
        return pointer_format in [
            DyldChainedPtrFormats.DYLD_CHAINED_PTR_64,
            DyldChainedPtrFormats.DYLD_CHAINED_PTR_64_OFFSET,
        ]


# noinspection PyPep8Naming
class ChainedFixupPointerOnDisk(ctypes.Union):
    """
    the ChainedFixupPointerOnDisk union from dyld MachOLoaded.h
    https://github.com/apple-opensource/dyld/blob/852.2/dyld3/MachOLoaded.h#L87-L141
    """

    generic64: Generic64
    arm64e: Arm64e

    _fields_ = [("generic64", Generic64), ("arm64e", Arm64e)]

    def isBind(self, pointer_format: DyldChainedPtrFormats) -> tuple[int, int] | None:
        """
        Port of ChainedFixupPointerOnDisk::isBind(uint16_t pointerFormat, uint32_t& bindOrdinal, int64_t& addend)
        https://github.com/apple-opensource/dyld/blob/852.2/dyld3/MachOLoaded.cpp#L1098-L1147
        Returns None if not a bind (so `if struct.isBind()` works),
        :return:
        """
        # pylint: disable=no-else-raise
        if Arm64e.check_valid_pointer_format(pointer_format):
            # https://github.com/apple-opensource/dyld/blob/852.2/dyld3/MachOLoaded.cpp#L1107-L1124
            if self.arm64e.authBind.bind:
                if self.arm64e.authBind.auth:
                    if pointer_format == DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E_USERLAND24:
                        return self.arm64e.authBind24.ordinal, 0
                    else:
                        return self.arm64e.authBind.ordinal, 0
                else:
                    if pointer_format == DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E_USERLAND24:
                        return self.arm64e.bind24.ordinal, self.arm64e.sign_extended_addend
                    else:
                        return self.arm64e.bind.ordinal, self.arm64e.sign_extended_addend
            else:
                return None
        elif Generic64.check_valid_pointer_format(pointer_format):
            # https://github.com/apple-opensource/dyld/blob/852.2/dyld3/MachOLoaded.cpp#L1126-L1132
            if self.generic64.bind.bind:
                return self.generic64.bind.ordinal, self.generic64.bind.addend
            else:
                return None

        else:
            raise NotImplementedError(f"Not yet supported pointer format {pointer_format}")

    def isRebase(
        self, pointer_format: DyldChainedPtrFormats, preferredLoadAddress: MemoryPointer
    ) -> MemoryPointer | None:
        """
        port of ChainedFixupPointerOnDisk::isRebase(
        uint16_t pointerFormat, uint64_t preferedLoadAddress, uint64_t& targetRuntimeOffset)
        https://github.com/apple-opensource/dyld/blob/852.2/dyld3/MachOLoaded.cpp#L1046-L1096
        :param pointer_format:
        :param preferredLoadAddress: I think that's just the requested base address
        :return:
        """
        # pylint: disable=no-else-raise
        if Arm64e.check_valid_pointer_format(pointer_format):
            # https://github.com/apple-opensource/dyld/blob/852.2/dyld3/MachOLoaded.cpp#L1049-L1067
            if self.arm64e.bind.bind:
                return False
            else:
                if self.arm64e.authRebase.auth:
                    return self.arm64e.authRebase.target
                else:
                    targetRuntimeOffset = self.arm64e.unpack_target
                    if pointer_format in [
                        DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E,
                        DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E_FIRMWARE,
                    ]:
                        targetRuntimeOffset -= preferredLoadAddress

                    return targetRuntimeOffset

        elif Generic64.check_valid_pointer_format(pointer_format):
            # https://github.com/apple-opensource/dyld/blob/852.2/dyld3/MachOLoaded.cpp#L1068-L1076
            rebase = self.generic64.rebase
            if rebase.bind:
                # Then this wasn't actually a rebase
                return False
            else:
                targetRuntimeOffset = rebase.unpackedTarget
                if pointer_format == DyldChainedPtrFormats.DYLD_CHAINED_PTR_64:
                    targetRuntimeOffset -= preferredLoadAddress
                return targetRuntimeOffset
        else:
            raise NotImplementedError(f"Not yet supported pointer format {pointer_format}")


# DYLD_CHAINED_PTR_BASE: Dict[DyldChainedPtrFormats, Type[ChainedFixupPointerOnDisk]] = {
#     DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E: Arm64e,
#     DyldChainedPtrFormats.DYLD_CHAINED_PTR_ARM64E_USERLAND24: Arm64e,
#     DyldChainedPtrFormats.DYLD_CHAINED_PTR_64: Generic64,
#     DyldChainedPtrFormats.DYLD_CHAINED_PTR_64_OFFSET: Generic64
#
# }


class DyldImportStruct(HelperStruct):
    """
    Meta Struct for the different kind of import structs and the fields they are all guaranteed to have
    """

    lib_ordinal: int
    weak_import: bool
    name_offset: FileOffset

    @staticmethod
    def get_struct(pointer: DyldImportFormats) -> type[DyldImportStruct]:
        return {
            DyldImportFormats.DYLD_CHAINED_IMPORT: dyld_chained_import,
            DyldImportFormats.DYLD_CHAINED_IMPORT_ADDEND64: dyld_chained_import_addend64,
            DyldImportFormats.DYLD_CHAINED_IMPORT_ADDEND: dyld_chained_import_addend,
        }[pointer]


# noinspection PyPep8Naming
class dyld_chained_import(DyldImportStruct):
    """
    Struct for symbol format DYLD_CHAINED_IMPORT

    """

    _fields_ = [
        ("lib_ordinal", c_uint32, 8),
        ("weak_import", c_uint32, 1),
        ("name_offset", c_uint32, 23),
    ]


# noinspection PyPep8Naming
class dyld_chained_import_addend(DyldImportStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L264-L271
    """

    addend: int
    _fields_ = [
        ("lib_ordinal", c_uint32, 8),
        ("weak_import", c_uint32, 1),
        ("name_offset", c_uint32, 23),
        ("addend", c_uint32),
    ]


# DYLD_CHAINED_IMPORT_ADDEND64
# noinspection PyPep8Naming
class dyld_chained_import_addend64(DyldImportStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L273-L281
    """

    addend: int
    _fields_ = [
        ("lib_ordinal", c_uint64, 16),
        ("weak_import", c_uint64, 1),
        ("reserved", c_uint64, 15),
        ("name_offset", c_uint64, 32),
        ("addend", c_uint64),
    ]


# noinspection PyPep8Naming
class dyld_chained_fixups_header(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L36-L46
    """

    fixups_version: int
    starts_offset: FileOffset
    imports_offset: FileOffset
    symbols_offset: FileOffset
    imports_count: int
    imports_format: DyldImportFormats
    symbols_format: int

    _fields_ = [
        ("fixups_version", c_uint32),
        # offset of dyld_chained_starts_in_image in chain_data
        ("starts_offset", c_uint32),
        # offset of imports table in chain_data
        ("imports_offset", c_uint32),
        # offset of symbol strings in chain_data
        ("symbols_offset", c_uint32),
        # number of imported symbol names
        ("imports_count", c_uint32),
        # DYLD_CHAINED_IMPORT*
        ("imports_format", c_uint32),
        # 0 => uncompressed, 1 => zlib compressed
        ("symbols_format", c_uint32),
    ]


# noinspection PyPep8Naming
class dyld_chained_starts_in_image(ctypes.Structure):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L48-L54
    """

    seg_count: int
    seg_info_offset: ctypes.Array

    _fields_ = [("seg_count", c_uint32), ("seg_info_offset", c_uint32 * 1)]


DYLD_CHAINED_PTR_START_NONE = 0xFFFF  # used in page_start[] to denote a page with no fixups
DYLD_CHAINED_PTR_START_MULTI = 0x8000  # used in page_start[] to denote a page which has multiple starts
DYLD_CHAINED_PTR_START_LAST = 0x8000  # used in chain_starts[] to denote last start in list for page


# noinspection PyPep8Naming
class dyld_chained_starts_in_segment(HelperStruct):
    """
    https://github.com/apple-opensource/dyld/blob/852.2/include/mach-o/fixup-chains.h#L56-L72
    """

    # size: int
    page_size: int
    _pointer_format: int
    segment_offset: int
    max_valid_pointer: int
    page_count: int
    page_start: int

    _fields_ = [
        # size of this (amount kernel needs to copy)
        ("_size", c_uint32),
        # 0x1000 or 0x4000
        ("page_size", c_uint16),
        # DYLD_CHAINED_PTR_*
        ("_pointer_format", c_uint16),
        # offset in memory to start of segment
        ("segment_offset", c_uint64),
        ("max_valid_pointer", c_uint32),
        ("page_count", c_uint16),
        ("page_start", c_uint16),
    ]

    @property
    def pointer_format(self) -> DyldChainedPtrFormats:
        return DyldChainedPtrFormats(self._pointer_format)


BIND_SPECIAL_DYLIB_WEAK_LOOKUP = -3



"""
CLE is an extensible binary loader. Its main goal is to take an executable program and any libraries it depends on and
produce an address space where that program is loaded and ready to run.

The primary interface to CLE is the Loader class.
"""

from __future__ import annotations

__version__ = "9.2.160"

from .address_translator import AT, AddressTranslator
from .backends import (
    ALL_BACKENDS,
    CGC,
    ELF,
    PE,
    TE,
    XBE,
    Apk,
    BackedCGC,
    Backend,
    Blob,
    Coff,
    ELFCore,
    ExceptionHandling,
    FunctionHint,
    FunctionHintSource,
    Hex,
    Jar,
    MachO,
    MetaELF,
    Minidump,
    NamedRegion,
    Region,
    Regions,
    Section,
    Segment,
    Soot,
    SRec,
    StaticArchive,
    Symbol,
    SymbolSubType,
    SymbolType,
    UefiFirmware,
    register_backend,
)
from .backends.externs import (
    ExternObject,
    ExternSegment,
    KernelObject,
    PointToPrecise,
    TOCRelocation,
)
from .backends.tls import (
    ELFCoreThreadManager,
    ELFThreadManager,
    InternalTLSRelocation,
    MinidumpThreadManager,
    PEThreadManager,
    ThreadManager,
    TLSObject,
)
from .errors import (
    CLECompatibilityError,
    CLEError,
    CLEFileNotFoundError,
    CLEInvalidBinaryError,
    CLEInvalidEncryptionError,
    CLEInvalidFileFormatError,
    CLEMemoryError,
    CLEOperationError,
    CLEUnknownFormatError,
)
from .gdb import GDB_SEARCH_PATH, convert_info_proc_maps, convert_info_sharedlibrary
from .loader import Loader
from .memory import Clemory, ClemoryBase, ClemoryReadOnlyView, ClemoryTranslator, ClemoryView
from .patched_stream import PatchedStream

__all__ = [
    "ALL_BACKENDS",
    "CGC",
    "ELF",
    "PE",
    "XBE",
    "Apk",
    "BackedCGC",
    "Backend",
    "Blob",
    "Coff",
    "ELFCore",
    "ExceptionHandling",
    "FunctionHint",
    "FunctionHintSource",
    "Hex",
    "SRec",
    "Jar",
    "MachO",
    "MetaELF",
    "Minidump",
    "NamedRegion",
    "Region",
    "Regions",
    "Section",
    "Segment",
    "Soot",
    "StaticArchive",
    "register_backend",
    "ExternObject",
    "ExternSegment",
    "KernelObject",
    "PointToPrecise",
    "TOCRelocation",
    "ELFCoreThreadManager",
    "ELFThreadManager",
    "InternalTLSRelocation",
    "MinidumpThreadManager",
    "PEThreadManager",
    "ThreadManager",
    "TLSObject",
    "CLECompatibilityError",
    "CLEError",
    "CLEFileNotFoundError",
    "CLEInvalidBinaryError",
    "CLEInvalidEncryptionError",
    "CLEInvalidFileFormatError",
    "CLEOperationError",
    "CLEUnknownFormatError",
    "CLEMemoryError",
    "convert_info_proc_maps",
    "convert_info_sharedlibrary",
    "GDB_SEARCH_PATH",
    "Loader",
    "Clemory",
    "ClemoryBase",
    "ClemoryReadOnlyView",
    "ClemoryView",
    "ClemoryTranslator",
    "PatchedStream",
    "AddressTranslator",
    "AT",
    "Symbol",
    "SymbolType",
    "SymbolSubType",
    "TE",
    "UefiFirmware",
]



"""
This module is used when parsing the Thread Local Storage of an ELF binary. It heavily uses the TLSArchInfo
namedtuple from archinfo.

ELF TLS is implemented based on the following documents:

    - https://www.uclibc.org/docs/tls.pdf
    - https://www.uclibc.org/docs/tls-ppc.txt
    - https://www.uclibc.org/docs/tls-ppc64.txt
    - https://www.linux-mips.org/wiki/NPTL
"""

from __future__ import annotations

from .tls_object import InternalTLSRelocation, ThreadManager, TLSObject

TLS_BLOCK_ALIGN = 0x10
TLS_TOTAL_HEAD_SIZE = 0x4000
TLS_HEAD_ALIGN = 0x10000


def roundup(val, to=TLS_BLOCK_ALIGN):
    return val - 1 + (to - ((val - 1) % to))


class ELFThreadManager(ThreadManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.used_data = 0

    def register_object(self, obj):
        if not super().register_object(obj):
            return False

        # only track tls_block_offset for modules registered before the first thread
        # ??? is this right
        if not self.threads:
            # tls_block_offset has these semantics: the thread pointer plus the block offset equals the address of the
            # module's TLS data in memory
            if self.arch.elf_tls.variant == 1:
                obj.tls_block_offset = self.used_data
            else:
                obj.tls_block_offset = -self.used_data - roundup(obj.tls_block_size)

            self.used_data += roundup(obj.tls_block_size)
        return True

    @property
    def _thread_cls(self):
        if self.arch.elf_tls.variant == 1:
            return ELFTLSObjectV1
        else:
            return ELFTLSObjectV2


class ELFTLSObject(TLSObject):
    def __init__(self, thread_manager: ELFThreadManager):
        super().__init__(loader=thread_manager.loader, arch=thread_manager.arch)
        self.tcb_offset: int = None
        self.dtv_offset: int = None
        self.tp_offset: int = None
        self.head_offset: int = None
        self._max_addr: int = None
        self.tlsoffsets = [obj.tls_block_offset for obj in thread_manager.modules]
        self.pic = True

        self._calculate_pointers(thread_manager.used_data, thread_manager.max_modules)

        # add backer for header
        self.memory.add_backer(self.head_offset, bytes(TLS_TOTAL_HEAD_SIZE))

        # add backer for dtv
        self.memory.add_backer(
            self.dtv_offset - 2 * self.arch.bytes,
            bytes(2 * self.arch.bytes * thread_manager.max_modules + 2 * self.arch.bytes),
        )

        # Set the appropriate pointers in the tcbhead
        for off in self.arch.elf_tls.head_offsets:
            self._drop_int(off + self.tcb_offset, self.tp_offset, True)
        for off in self.arch.elf_tls.dtv_offsets:
            self._drop_int(off + self.tcb_offset, self.dtv_offset, True)
        for off in self.arch.elf_tls.pthread_offsets:
            self._drop_int(off + self.tcb_offset, self.tp_offset, True)  # ?????

        # tid. feel free to move this code wherever
        # this only matters if you're not running the libc initializers... hm.
        # tid = len(thread_manager.threads) + 1
        # if self.arch.name == 'AMD64':
        #    self._drop_int(self.tcb_offset + 0x2d0, tid, False, size=4)

        # Set up the DTV
        # at dtv[-1] there's capacity, at dtv[0] there's count (technically generation number?)
        self._drop_int(self.dtv_offset - 2 * self.arch.bytes, thread_manager.max_modules - 1)
        self._drop_int(self.dtv_offset, len(thread_manager.modules))

        # set up each module
        for obj in thread_manager.modules:
            # dtv entry
            dtv_entry_offset = self.dtv_offset + 2 * self.arch.bytes * obj.tls_module_id
            self._drop_int(
                dtv_entry_offset, self.tp_offset + obj.tls_block_offset + self.arch.elf_tls.dtv_entry_offset, True
            )
            self._drop_int(dtv_entry_offset + self.arch.bytes, 1)

            # initialization image
            image = thread_manager.initialization_image(obj)
            if image is None:
                continue
            self.memory.add_backer(self.tp_offset + obj.tls_block_offset, image)

    def _calculate_pointers(self, used_data, max_modules):
        raise NotImplementedError

    def _drop_int(self, offset, num, needs_relocation=False, **kwargs):
        if needs_relocation:
            self.relocs.append(InternalTLSRelocation(num, offset, self))
        self.memory.pack_word(offset, num, **kwargs)

    @property
    def thread_pointer(self):
        """
        The thread pointer. This is a technical term that refers to a specific location in the TLS segment.
        """
        return self.mapped_base + self.tp_offset

    @property
    def user_thread_pointer(self):
        """
        The thread pointer that is exported to the user
        """
        return self.thread_pointer + self.arch.elf_tls.tp_offset

    @property
    def max_addr(self):
        return self.mapped_base + self._max_addr

    def get_addr(self, module_id, offset):
        """
        basically ``__tls_get_addr``.
        """
        return self.memory.unpack_word(self.dtv_offset + module_id * self.arch.bytes) + offset


class ELFTLSObjectV1(ELFTLSObject):
    # variant 1: memory is laid out like so:
    # [header][module data]
    #         ^ thread pointer
    def _calculate_pointers(self, used_data, max_modules):
        self.tcb_offset = TLS_TOTAL_HEAD_SIZE - self.arch.elf_tls.tcbhead_size
        # CRITICAL DIFFERENCE FROM THE DOC - variant 1 seems to expect the thread pointer points to the end of the TCB
        self.tp_offset = TLS_TOTAL_HEAD_SIZE
        self.dtv_offset = TLS_TOTAL_HEAD_SIZE + used_data + 2 * self.arch.bytes
        self.head_offset = 0  # ^^ that's the point of this field
        self._max_addr = self.dtv_offset + 2 * self.arch.bytes * max_modules - 1


class ELFTLSObjectV2(ELFTLSObject):
    # variant 2: memory is laid out like so:
    # [module data][header]
    #              ^ thread pointer
    def _calculate_pointers(self, used_data, max_modules):
        self.tcb_offset = roundup(used_data, TLS_HEAD_ALIGN)
        self.tp_offset = roundup(used_data, TLS_HEAD_ALIGN)
        self.dtv_offset = self.tp_offset + TLS_TOTAL_HEAD_SIZE + 2 * self.arch.bytes
        self.head_offset = self.tp_offset
        self._max_addr = self.dtv_offset + 2 * self.arch.bytes * max_modules - 1



from __future__ import annotations

import io
import logging
import mmap
from dataclasses import dataclass
from functools import singledispatchmethod
from uuid import UUID

import archinfo

from cle.errors import CLEError, CLEUnknownFormatError

from . import Backend, register_backend
from .pe import PE
from .te import TE

try:
    import uefi_firmware
except ImportError:
    uefi_firmware = None

log = logging.getLogger(__name__)


class UefiDriverLoadError(Exception):
    """
    This error is raised (and caught internally) if the data contained in the UEFI entity tree doesn't make sense.
    """


class UefiFirmware(Backend):
    """
    A UEFI firmware blob loader. Support is provided by the ``uefi_firmware`` package.
    """

    is_default = True

    @classmethod
    def _to_bytes(cls, fileobj: io.IOBase):
        try:
            fileno = fileobj.fileno()
        except io.UnsupportedOperation:
            pass
        else:
            return mmap.mmap(fileno, 0, access=mmap.ACCESS_READ)

        if isinstance(fileobj, io.BytesIO):
            return fileobj.getbuffer()

        # fuck it, we'll do it live
        fileobj.seek(0)
        return fileobj.read()

    @classmethod
    def is_compatible(cls, stream):
        if uefi_firmware is None:
            return False

        buffer = cls._to_bytes(stream)
        parser = uefi_firmware.AutoParser(buffer)
        return parser.type() != "unknown"

    def __init__(self, *args, **kwargs) -> None:
        if uefi_firmware is None:
            raise CLEError("run `pip install uefi_firmware==1.10` to load UEFI firmware")
        super().__init__(*args, **kwargs)

        # hack: we are using a loader internal method in a non-kosher way which will cause our children to be
        # marked as the main binary if we are also the main binary
        # work around this by setting ourself here:
        if self.loader._main_object is None:
            self.loader._main_object = self

        self._drivers: dict[UUID, UefiModuleMixin] = {}
        self._drivers_pending: dict[UUID, UefiModulePending] = {}
        self._current_file: UUID | None = None

        self.set_arch(archinfo.arch_from_id("x86_64"))  # TODO: ???

        buffer = self._to_bytes(self._binary_stream)
        parser = uefi_firmware.AutoParser(buffer)
        firmware = parser.parse()
        self._load(firmware)

        while self._drivers_pending:
            uuid, pending = self._drivers_pending.popitem()
            try:
                child = pending.build(self, uuid)
            except UefiDriverLoadError as e:
                log.warning("Failed to load %s: %s", uuid, e.args[0])
            else:
                self._drivers[uuid] = child
                child.parent_object = self
                self.child_objects.append(child)

        if self.child_objects:
            self._arch = self.child_objects[0].arch
        else:
            log.warning("Loaded empty UEFI firmware?")
        self.has_memory = False
        self.pic = True

        # hack pt. 2
        if self.loader._main_object is self:
            self.loader._main_object = None

    @singledispatchmethod
    def _load(self, uefi_obj):  # pylint: disable=no-self-use
        raise CLEUnknownFormatError(f"Can't load firmware object: {uefi_obj}")

    if uefi_firmware is not None:

        @_load.register
        def _load_generic(self, uefi_obj: uefi_firmware.FirmwareObject):
            for obj in uefi_obj.objects:
                self._load(obj)

        @_load.register
        def _load_none(self, uefi_obj: None):
            pass

        @_load.register
        def _load_firmwarefile(self, uefi_obj: uefi_firmware.uefi.FirmwareFile):
            old_uuid = self._current_file
            if uefi_obj.type == 7:  # driver
                uuid = UUID(bytes=uefi_obj.guid)
                self._drivers_pending[uuid] = UefiModulePending()
                self._current_file = uuid
            self._load_generic(uefi_obj)
            self._current_file = old_uuid

        @_load.register
        def _load_firmwarefilesection(self, uefi_obj: uefi_firmware.uefi.FirmwareFileSystemSection):
            pending = self._drivers_pending.get(self._current_file, None)
            if pending is not None:
                if uefi_obj.type == 16:  # pe32 image
                    pending.pe_image = uefi_obj.content
                elif uefi_obj.type == 18:  # te image
                    pending.te_image = uefi_obj.content
                elif uefi_obj.type == 21:  # user interface name
                    pending.name = uefi_obj.content.decode("utf-16").strip("\0")
            self._load_generic(uefi_obj)


@dataclass
class UefiModulePending:
    """
    A worklist entry for the UEFI firmware loader.
    """

    name: str | None = None
    pe_image: bytes | None = None
    te_image: bytes | None = None
    # version
    # dependencies

    def build(self, parent: UefiFirmware, guid: UUID) -> UefiModuleMixin:
        count = (self.pe_image is not None) + (self.te_image is not None)
        if count > 1:
            raise UefiDriverLoadError("Multiple image sections")
        cls: type[UefiModuleMixin]
        if self.pe_image is not None:
            cls = UefiPE
            data = self.pe_image
        elif self.te_image is not None:
            cls = UefiTE
            data = self.te_image
        else:
            raise UefiDriverLoadError("Missing PE or TE image section")
        return cls(None, io.BytesIO(data), is_main_bin=False, loader=parent.loader, name=self.name, guid=guid)


class UefiModuleMixin(Backend):
    """
    A mixin to make other kinds of backends load as UEFI modules.
    """

    def __init__(self, *args, guid: UUID, name: str | None, **kwargs):
        super().__init__(*args, **kwargs)
        self.guid = guid
        self.user_interface_name = name

        if self.linked_base == 0:
            self.pic = True

    def __repr__(self):
        return (
            f"<{type(self).__name__} Object "
            f'{self.guid}{f" {self.user_interface_name}" if self.user_interface_name else ""}, '
            f"maps [{self.min_addr:#x}:{self.max_addr:#x}]>"
        )


class UefiPE(UefiModuleMixin, PE):
    """
    A PE file contained in a UEFI image.
    """


class UefiTE(UefiModuleMixin, TE):
    """
    A TE file contained in a UEFI image.
    """


register_backend("uefi", UefiFirmware)



"""Relocations for AARCH64

Reference: https://github.com/ARM-software/abi-aa/blob/main/aaelf64/aaelf64.rst#relocation
"""

from __future__ import annotations

import logging

from .elfreloc import ELFReloc
from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericIRelativeReloc,
    GenericJumpslotReloc,
    GenericPCRelativeAddendReloc,
    GenericRelativeReloc,
    GenericTLSDescriptorReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
)

log = logging.getLogger(name=__name__)


class R_AARCH64_ABS64(GenericAbsoluteAddendReloc):
    pass


class R_AARCH64_COPY(GenericCopyReloc):
    pass


class R_AARCH64_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_AARCH64_JUMP_SLOT(GenericJumpslotReloc):
    pass


class R_AARCH64_RELATIVE(GenericRelativeReloc):
    pass


class R_AARCH64_IRELATIVE(GenericIRelativeReloc):
    pass


class R_AARCH64_TLS_DTPREL(GenericTLSDoffsetReloc):
    pass


class R_AARCH64_TLS_DTPMOD(GenericTLSModIdReloc):
    pass


class R_AARCH64_TLS_TPREL(GenericTLSOffsetReloc):
    pass


class R_AARCH64_TLSDESC(GenericTLSDescriptorReloc):
    RESOLVER_ADDR = 0xFFFF_FFFF_FFFF_FE00


class R_AARCH64_PREL32(GenericPCRelativeAddendReloc):
    """
    Relocation Type: 261
    Calculation: (S + A - P)
    """


class R_AARCH64_JUMP26(ELFReloc):
    """
    Relocation Type: 282
    Calculation: (S + A - P)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr
        return S + A - P

    def relocate(self):
        if not self.resolved:
            return False
        if not ((-(2**27)) <= self.value and self.value < (2**27)):
            log.warning("relocation out of range")
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111100000000000000000000000000
        imm = self.value >> 2 & 0b0000_0011_1111_1111_1111_1111_1111_1111  # [27:2] of the value
        self.owner.memory.pack_word(self.relative_addr, instr | imm, size=4)
        return True


class R_AARCH64_CALL26(ELFReloc):
    """
    Relocation Type: 283
    Calculation: (S + A - P)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr
        return S + A - P

    def relocate(self):
        if not self.resolved:
            return False
        if not ((-(2**27)) <= self.value and self.value < (2**27)):
            log.warning("relocation out of range")
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111100000000000000000000000000
        imm = self.value >> 2 & 0b0000_0011_1111_1111_1111_1111_1111_1111  # [27:2] of the value
        self.owner.memory.pack_word(self.relative_addr, instr | imm, size=4)
        return True


class R_AARCH64_ADR_PREL_PG_HI21(ELFReloc):
    """
    Relocation Type: 275
    Calculation: Page(S + A) - Page(P)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr
        return ((S + A) & ~0xFFF) - (P & ~0xFFF)

    def relocate(self):
        if not self.resolved:
            return False
        if not ((-(2**32)) <= self.value and self.value < (2**32)):
            log.warning("relocation out of range")
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b10011111000000000000000000011111
        imm = self.value >> 12 & 0x1FFFFF
        immlo = imm & 0b11
        immhi = imm >> 2
        self.owner.memory.pack_word(self.relative_addr, instr | (immhi << 5) | (immlo << 29), size=4)
        return True


class R_AARCH64_ADD_ABS_LO12_NC(ELFReloc):
    """
    Relocation Type: 277
    Calculation: (S + A)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        return S + A

    def relocate(self):
        if not self.resolved:
            return False
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111111110000000000001111111111
        imm = self.value & 0b1111_1111_1111
        self.owner.memory.pack_word(self.relative_addr, instr | (imm << 10), size=4)
        return True


class R_AARCH64_LDST8_ABS_LO12_NC(ELFReloc):
    """
    Relocation Type: 278
    Calculation: (S + A)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        return S + A

    def relocate(self):
        if not self.resolved:
            return False
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111111110000000000001111111111
        imm = self.value & 0b1111_1111_1111
        self.owner.memory.pack_word(self.relative_addr, instr | (imm << 10), size=4)
        return True


class R_AARCH64_LDST16_ABS_LO12_NC(ELFReloc):
    """
    Relocation Type: 284
    Calculation: (S + A)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        return S + A

    def relocate(self):
        if not self.resolved:
            return False
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111111110000000000001111111111
        imm = (self.value & 0b1111_1111_1111) >> 1
        self.owner.memory.pack_word(self.relative_addr, instr | (imm << 10), size=4)
        return True


class R_AARCH64_LDST32_ABS_LO12_NC(ELFReloc):
    """
    Relocation Type: 285
    Calculation: (S + A)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        return S + A

    def relocate(self):
        if not self.resolved:
            return False
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111111110000000000001111111111
        imm = (self.value & 0b1111_1111_1111) >> 2
        self.owner.memory.pack_word(self.relative_addr, instr | (imm << 10), size=4)
        return True


class R_AARCH64_LDST64_ABS_LO12_NC(ELFReloc):
    """
    Relocation Type: 286
    Calculation: (S + A)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        return S + A

    def relocate(self):
        if not self.resolved:
            return False
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111111110000000000001111111111
        imm = (self.value & 0b1111_1111_1111) >> 3
        self.owner.memory.pack_word(self.relative_addr, instr | (imm << 10), size=4)
        return True


class R_AARCH64_LDST128_ABS_LO12_NC(ELFReloc):
    """
    Relocation Type: 299
    Calculation: (S + A)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        return S + A

    def relocate(self):
        if not self.resolved:
            return False
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111111110000000000001111111111
        imm = (self.value & 0b1111_1111_1111) >> 4
        self.owner.memory.pack_word(self.relative_addr, instr | (imm << 10), size=4)
        return True


relocation_table_arm64 = {
    257: R_AARCH64_ABS64,
    258: R_AARCH64_COPY,
    # 259: R_AARCH64_ABS16,
    # 260: R_AARCH64_PREL64,
    261: R_AARCH64_PREL32,
    # 262: R_AARCH64_PREL16,
    # 263: R_AARCH64_MOVW_UABS_G0,
    # 264: R_AARCH64_MOVW_UABS_G0_NC,
    # 265: R_AARCH64_MOVW_UABS_G1,
    # 266: R_AARCH64_MOVW_UABS_G1_NC,
    # 267: R_AARCH64_MOVW_UABS_G2,
    # 268: R_AARCH64_MOVW_UABS_G2_NC,
    # 269: R_AARCH64_MOVW_UABS_G3,
    # 270: R_AARCH64_MOVW_SABS_G0,
    # 271: R_AARCH64_MOVW_SABS_G1,
    # 272: R_AARCH64_MOVW_SABS_G2,
    # 273: R_AARCH64_LD_PREL_LO19,
    # 274: R_AARCH64_ADR_PREL_LO21,
    275: R_AARCH64_ADR_PREL_PG_HI21,
    # 276: R_AARCH64_ADR_PREL_PG_HI21_NC,
    277: R_AARCH64_ADD_ABS_LO12_NC,
    278: R_AARCH64_LDST8_ABS_LO12_NC,
    # 279: R_AARCH64_TSTBR14,
    # 280: R_AARCH64_CONDBR19,
    282: R_AARCH64_JUMP26,
    283: R_AARCH64_CALL26,
    284: R_AARCH64_LDST16_ABS_LO12_NC,
    285: R_AARCH64_LDST32_ABS_LO12_NC,
    286: R_AARCH64_LDST64_ABS_LO12_NC,
    # 287: R_AARCH64_MOVW_PREL_G0,
    # 288: R_AARCH64_MOVW_PREL_G0_NC,
    # 289: R_AARCH64_MOVW_PREL_G1,
    # 290: R_AARCH64_MOVW_PREL_G1_NC,
    # 291: R_AARCH64_MOVW_PREL_G2,
    # 292: R_AARCH64_MOVW_PREL_G2_NC,
    # 293: R_AARCH64_MOVW_PREL_G3,
    299: R_AARCH64_LDST128_ABS_LO12_NC,
    # 300: R_AARCH64_MOVW_GOTOFF_G0,
    # 301: R_AARCH64_MOVW_GOTOFF_G0_NC,
    # 302: R_AARCH64_MOVW_GOTOFF_G1,
    # 303: R_AARCH64_MOVW_GOTOFF_G1_NC,
    # 304: R_AARCH64_MOVW_GOTOFF_G2,
    # 305: R_AARCH64_MOVW_GOTOFF_G2_NC,
    # 306: R_AARCH64_MOVW_GOTOFF_G3,
    # 307: R_AARCH64_GOTREL64,
    # 308: R_AARCH64_GOTREL32,
    # 309: R_AARCH64_GOT_LD_PREL19,
    # 310: R_AARCH64_LD64_GOTOFF_LO15,
    # 311: R_AARCH64_ADR_GOT_PAGE,
    # 312: R_AARCH64_LD64_GOT_LO12_NC,
    # 313: R_AARCH64_LD64_GOTPAGE_LO15,
    # 314: R_AARCH64_PLT32,
    # 315: R_AARCH64_GOTPCREL32,
    # 512: R_AARCH64_TLSGD_ADR_PREL21,
    # 513: R_AARCH64_TLSGD_ADR_PAGE21,
    # 514: R_AARCH64_TLSGD_ADD_LO12_NC,
    # 515: R_AARCH64_TLSGD_MOVW_G1,
    # 516: R_AARCH64_TLSGD_MOVW_G0_NC,
    # 517: R_AARCH64_TLSLD_ADR_PREL21,
    # 518: R_AARCH64_TLSLD_ADR_PAGE21,
    # 519: R_AARCH64_TLSLD_ADD_LO12_NC,
    # 520: R_AARCH64_TLSLD_MOVW_G1,
    # 521: R_AARCH64_TLSLD_MOVW_G0_NC,
    # 522: R_AARCH64_TLSLD_LD_PREL19,
    # 523: R_AARCH64_TLSLD_MOVW_DTPREL_G2,
    # 524: R_AARCH64_TLSLD_MOVW_DTPREL_G1,
    # 525: R_AARCH64_TLSLD_MOVW_DTPREL_G1_NC,
    # 526: R_AARCH64_TLSLD_MOVW_DTPREL_G0,
    # 527: R_AARCH64_TLSLD_MOVW_DTPREL_G0_NC,
    # 528: R_AARCH64_TLSLD_ADD_DTPREL_HI12,
    # 529: R_AARCH64_TLSLD_ADD_DTPREL_LO12,
    # 530: R_AARCH64_TLSLD_ADD_DTPREL_LO12_NC,
    # 531: R_AARCH64_TLSLD_LDST8_DTPREL_LO12,
    # 532: R_AARCH64_TLSLD_LDST8_DTPREL_LO12_NC,
    # 533: R_AARCH64_TLSLD_LDST16_DTPREL_LO12,
    # 534: R_AARCH64_TLSLD_LDST16_DTPREL_LO12_NC,
    # 535: R_AARCH64_TLSLD_LDST32_DTPREL_LO12,
    # 536: R_AARCH64_TLSLD_LDST32_DTPREL_LO12_NC,
    # 537: R_AARCH64_TLSLD_LDST64_DTPREL_LO12,
    # 538: R_AARCH64_TLSLD_LDST64_DTPREL_LO12_NC,
    # 539: R_AARCH64_TLSIE_MOVW_GOTTPREL_G1,
    # 540: R_AARCH64_TLSIE_MOVW_GOTTPREL_G0_NC,
    # 541: R_AARCH64_TLSIE_ADR_GOTTPREL_PAGE21,
    # 542: R_AARCH64_TLSIE_LD64_GOTTPREL_LO12_NC,
    # 543: R_AARCH64_TLSIE_LD_GOTTPREL_PREL19,
    # 544: R_AARCH64_TLSLE_MOVW_TPREL_G2,
    # 545: R_AARCH64_TLSLE_MOVW_TPREL_G1,
    # 546: R_AARCH64_TLSLE_MOVW_TPREL_G1_NC,
    # 547: R_AARCH64_TLSLE_MOVW_TPREL_G0,
    # 548: R_AARCH64_TLSLE_MOVW_TPREL_G0_NC,
    # 549: R_AARCH64_TLSLE_ADD_TPREL_HI12,
    # 550: R_AARCH64_TLSLE_ADD_TPREL_LO12,
    # 551: R_AARCH64_TLSLE_ADD_TPREL_LO12_NC,
    # 552: R_AARCH64_TLSLE_LDST8_TPREL_LO12,
    # 553: R_AARCH64_TLSLE_LDST8_TPREL_LO12_NC,
    # 554: R_AARCH64_TLSLE_LDST16_TPREL_LO12,
    # 555: R_AARCH64_TLSLE_LDST16_TPREL_LO12_NC,
    # 556: R_AARCH64_TLSLE_LDST32_TPREL_LO12,
    # 557: R_AARCH64_TLSLE_LDST32_TPREL_LO12_NC,
    # 558: R_AARCH64_TLSLE_LDST64_TPREL_LO12,
    # 559: R_AARCH64_TLSLE_LDST64_TPREL_LO12_NC,
    # 560: R_AARCH64_TLSDESC_LD_PREL19,
    # 561: R_AARCH64_TLSDESC_ADR_PREL21,
    # 562: R_AARCH64_TLSDESC_ADR_PAGE21,
    # 563: R_AARCH64_TLSDESC_LD64_LO12,
    # 564: R_AARCH64_TLSDESC_ADD_LO12,
    # 565: R_AARCH64_TLSDESC_OFF_G1,
    # 566: R_AARCH64_TLSDESC_OFF_G0_NC,
    # 567: R_AARCH64_TLSDESC_LDR,
    # 568: R_AARCH64_TLSDESC_ADD,
    # 569: R_AARCH64_TLSDESC_CALL,
    # 570: R_AARCH64_TLSLE_LDST128_TPREL_LO12,
    # 571: R_AARCH64_TLSLE_LDST128_TPREL_LO12_NC,
    # 572: R_AARCH64_TLSLD_LDST128_DTPREL_LO12,
    # 573: R_AARCH64_TLSLD_LDST128_DTPREL_LO12_NC,
    # 580: R_AARCH64_AUTH_ABS64,
    1024: R_AARCH64_COPY,
    1025: R_AARCH64_GLOB_DAT,
    1026: R_AARCH64_JUMP_SLOT,
    1027: R_AARCH64_RELATIVE,
    1028: R_AARCH64_TLS_DTPMOD,  # R_AARCH64_TLS_IMPDEF1
    1029: R_AARCH64_TLS_DTPREL,  # R_AARCH64_TLS_IMPDEF2
    1030: R_AARCH64_TLS_TPREL,
    1031: R_AARCH64_TLSDESC,
    1032: R_AARCH64_IRELATIVE,
    # 1040: R_AARCH64_AUTH_ABS64,
    # 1041: R_AARCH64_AUTHELATIVE,
}

__all__ = ("relocation_table_arm64",)



from __future__ import annotations

import logging

from .arm import relocation_table_arm
from .generic import relocation_table_generic
from .mips import relocation_table_mips
from .riscv import relocation_table_riscv

ALL_RELOCATIONS = {
    "AMD64": relocation_table_generic,
    "arm": relocation_table_generic | relocation_table_arm,
    "X86": relocation_table_generic,
    "mips": relocation_table_generic | relocation_table_mips,
    "RISCV": relocation_table_generic | relocation_table_riscv,
}

log = logging.getLogger(name=__name__)
complaint_log = set()


def get_relocation(arch, r_type):
    if r_type == 0:
        return None
    try:
        return ALL_RELOCATIONS[arch][r_type]
    except KeyError:
        if (arch, r_type) not in complaint_log:
            complaint_log.add((arch, r_type))
            log.warning("Unknown reloc %d on %s", r_type, arch)
        return None



from __future__ import annotations

import itertools
import logging
import os
import platform
import sys
from collections import OrderedDict
from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Literal,
    TypeVar,
)

import archinfo
from archinfo.arch_soot import ArchSoot

from cle import Symbol
from cle.address_translator import AT
from cle.errors import CLECompatibilityError, CLEError, CLEFileNotFoundError, CLEOperationError
from cle.memory import Clemory, ClemoryReadOnlyView
from cle.utils import ALIGN_UP, key_bisect_floor_key, key_bisect_insort_right, stream_or_path

from .backends import ALL_BACKENDS, ELF, PE, Backend, Blob, Coff, ELFCore, MetaELF, Minidump
from .backends.externs import ExternObject, KernelObject
from .backends.tls import (
    ELFCoreThreadManager,
    ELFThreadManager,
    MinidumpThreadManager,
    PEThreadManager,
    ThreadManager,
    TLSObject,
)

__all__ = ("Loader",)

log = logging.getLogger(name=__name__)

if TYPE_CHECKING:
    from .backends import Region, Section, Segment
    from .backends.relocation import Relocation

T = TypeVar("T")


class Loader:
    """
    The loader loads all the objects and exports an abstraction of the memory of the process. What you see here is an
    address space with loaded and rebased binaries.
    """

    def __init__(
        self,
        main_binary: str | BinaryIO | Path | Backend,
        auto_load_libs: bool = True,
        concrete_target=None,
        force_load_libs: Iterable[str | BinaryIO | Path] = (),
        skip_libs: Iterable[str] = (),
        main_opts: dict[str, Any] | None = None,
        lib_opts: dict[str, dict[str, Any]] | None = None,
        ld_path: Iterable[str | Path] = (),
        use_system_libs: bool = True,
        ignore_import_version_numbers: bool = True,
        case_insensitive: bool = False,
        rebase_granularity: int = 0x100000,
        except_missing_libs: bool = False,
        aslr: bool = False,
        perform_relocations: bool = True,
        load_debug_info: bool = False,
        page_size: int = 0x1,
        preload_libs: Iterable[str | BinaryIO | Path] = (),
        arch: archinfo.Arch | str | None = None,
    ):
        """
        :param main_binary:         The path to the main binary you're loading, or a file-like object with the binary
                                    in it.

        :param auto_load_libs:      Whether to automatically load shared libraries that loaded objects depend on.
        :param load_debug_info:     Whether to automatically parse DWARF data and search for debug symbol files.
        :param concrete_target:     Whether to instantiate a concrete target for a concrete execution of the process.
                                    if this is the case we will need to instantiate a SimConcreteEngine that wraps the
                                    ConcreteTarget provided by the user.
        :param force_load_libs:     A list of libraries to load regardless of if they're required by a loaded object.
        :param skip_libs:           A list of libraries to never load, even if they're required by a loaded object.
        :param main_opts:           A dictionary of options to be used loading the main binary.
        :param lib_opts:            A dictionary mapping library names to the dictionaries of options to be used when
                                    loading them.
        :param ld_path:             A list of paths in which we can search for shared libraries.
        :param use_system_libs:     Whether or not to search the system load path for requested libraries. Default True.
        :param ignore_import_version_numbers:
                                    Whether libraries with different version numbers in the filename will be considered
                                    equivalent, for example libc.so.6 and libc.so.0
        :param case_insensitive:    If this is set to True, filesystem loads will be done case-insensitively regardless
                                    of the case-sensitivity of the underlying filesystem.
        :param rebase_granularity:  The alignment to use for rebasing shared objects
        :param except_missing_libs: Throw an exception when a shared library can't be found.
        :param aslr:                Load libraries in symbolic address space. Do not use this option.
        :param page_size:           The granularity with which data is mapped into memory. Set to 0x1000 if you are
                                    working in an environment where data will always be memory mapped in a page-graunlar
                                    way.
        :param preload_libs:        Similar to `force_load_libs` but will provide for symbol resolution, with precedence
                                    over any dependencies.

        :ivar memory:               The loaded, rebased, and relocated memory of the program.
        :vartype memory:            cle.memory.Clemory
        :ivar main_object:          The object representing the main binary (i.e., the executable).
        :ivar shared_objects:       A dictionary mapping loaded library names to the objects representing them.
        :ivar all_objects:          A list containing representations of all the different objects loaded.
        :ivar requested_names:      A set containing the names of all the different shared libraries that were marked as
                                    a dependency by somebody.
        :ivar initial_load_objects: A list of all the objects that were loaded as a result of the initial load request.

        When reference is made to a dictionary of options, it requires a dictionary with zero or more of the following
        keys:

        - backend :             "elf", "pe", "mach-o", "blob" : which loader backend to use
        - arch :                The archinfo.Arch object to use for the binary
        - base_addr :           The address to rebase the object at
        - entry_point :         The entry point to use for the object

        More keys are defined on a per-backend basis.
        """
        if hasattr(main_binary, "seek") and hasattr(main_binary, "read"):
            self._main_binary_path = None
            self._main_binary_stream = main_binary
        elif isinstance(main_binary, Path):
            self._main_binary_path = str(main_binary.resolve())
            self._main_binary_stream = None
        else:
            self._main_binary_path = os.path.realpath(str(main_binary))
            self._main_binary_stream = None

        # whether we are presently in the middle of a load cycle
        self._juggling = False

        # auto_load_libs doesn't make any sense if we have a concrete target.
        if concrete_target:
            auto_load_libs = False

        self._auto_load_libs = auto_load_libs
        self._load_debug_info = load_debug_info
        self._satisfied_deps: dict[str, Literal[False] | Backend] = {x: False for x in skip_libs}
        self._main_opts = {} if main_opts is None else main_opts
        self._lib_opts = {} if lib_opts is None else lib_opts
        self._custom_ld_path = [ld_path] if isinstance(ld_path, str) else ld_path
        force_load_libs = [force_load_libs] if isinstance(force_load_libs, str) else force_load_libs
        preload_libs = [preload_libs] if isinstance(preload_libs, str) else preload_libs
        self._use_system_libs = use_system_libs
        self._ignore_import_version_numbers = ignore_import_version_numbers
        self._case_insensitive = case_insensitive
        self._rebase_granularity = rebase_granularity
        self._except_missing_libs = except_missing_libs
        self._relocated_objects = set()
        self._perform_relocations = perform_relocations

        # case insensitivity setup
        if sys.platform == "win32":  # TODO: a real check for case insensitive filesystems
            if self._main_binary_path:
                self._main_binary_path = self._main_binary_path.lower()
            force_load_libs = [x.lower() if isinstance(x, str) else x for x in force_load_libs]
            for x in list(self._satisfied_deps):
                self._satisfied_deps[x.lower()] = self._satisfied_deps[x]
            for x in list(self._lib_opts):
                self._lib_opts[x.lower()] = self._lib_opts[x]
            self._custom_ld_path = [x.lower() for x in self._custom_ld_path]

        self.aslr = aslr
        self.page_size = page_size
        self._memory = None
        self._original_main_object = None
        self._main_object = None
        self._tls = None
        self._kernel_object: KernelObject | None = None
        self._extern_object: ExternObject | None = None
        self.shared_objects = OrderedDict()
        self.all_objects: list[Backend] = []
        self.requested_names = set()
        if arch is not None:
            self._main_opts.update({"arch": arch})
        self.preload_libs = []
        self.initial_load_objects = self._internal_load(
            main_binary, *preload_libs, *force_load_libs, preloading=(main_binary, *preload_libs)
        )

        # cache
        self._last_object = None
        self._memory_ro_view = None

        if self._extern_object and self._extern_object._warned_data_import:
            log.warning(
                'For more information about "Symbol was allocated without a known size",'
                "see https://docs.angr.io/extending-angr/environment#simdata"
            )

    # Basic functions and properties

    @property
    def main_object(self) -> Backend:
        result = self._main_object
        if result is None:
            raise ValueError("Cannot access main_object before loading is complete")
        return result

    @property
    def original_main_object(self) -> Backend:
        result = self._original_main_object
        if result is None:
            raise ValueError("Cannot access original_main_object before loading is complete")
        return result

    @property
    def memory(self) -> Clemory:
        result = self._memory
        if result is None:
            raise ValueError("Cannot access memory before loading is complete")
        return result

    @property
    def memory_ro_view(self) -> ClemoryReadOnlyView | None:
        if self._memory is None:
            # it is intentional to check if self._memory is configured when memory_ro_view is accessed.
            # memory_ro_view is only set up after gen_ro_memview() is called.
            raise ValueError("Cannot access memory_ro_view before loading is complete")
        return self._memory_ro_view

    @property
    def tls(self) -> ThreadManager:
        result = self._tls
        if result is None:
            raise ValueError("Cannot access tls before loading is complete")
        return result

    def close(self):
        log.warning("You don't need to close the loader anymore :)")

    def __repr__(self):
        if self._main_binary_stream is None and self._main_binary_path is not None:
            return f"<Loaded {os.path.basename(self._main_binary_path)}, maps [{self.min_addr:#x}:{self.max_addr:#x}]>"
        else:
            return f"<Loaded from stream, maps [{self.min_addr:#x}:{self.max_addr:#x}]>"

    @property
    def max_addr(self) -> int:
        """
        The maximum address loaded as part of any loaded object (i.e., the whole address space).
        """
        return self.all_objects[-1].max_addr

    @property
    def min_addr(self) -> int:
        """
        The minimum address loaded as part of any loaded object (i.e., the whole address space).
        """
        return self.all_objects[0].min_addr

    @property
    def initializers(self) -> list[int]:
        """
        Return a list of all the initializers that should be run before execution reaches the entry point, in the order
        they should be run.
        """
        return sum((x.initializers for x in self.all_objects), [])

    @property
    def finalizers(self) -> list[int]:
        """
        Return a list of all the finalizers that should be run before the program exits.
        I'm not sure what order they should be run in.
        """
        return sum((x.finalizers for x in self.all_objects), [])

    @property
    def linux_loader_object(self) -> Backend | None:
        """
        If the linux dynamic loader is present in memory, return it
        """
        for obj in self.all_objects:
            if obj.provides is None:
                continue
            if self._is_linux_loader_name(obj.provides) is True:
                return obj
        return None

    @property
    def elfcore_object(self) -> ELFCore | None:
        """
        If a corefile was loaded, this returns the actual core object instead of the main binary
        """
        for obj in self.all_objects:
            if isinstance(obj, ELFCore):
                return obj
        return None

    @property
    def extern_object(self) -> ExternObject:
        """
        Return the extern object used to provide addresses to unresolved symbols and angr internals.

        Accessing this property will load this object into memory if it was not previously present.

        proposed model for how multiple extern objects should work:

        1) extern objects are a linked list. the one in loader._extern_object is the head of the list
        2) each round of explicit loads generates a new extern object if it has unresolved dependencies. this object
            has exactly the size necessary to hold all its exports.
        3) All requests for size are passed down the chain until they reach an object which has the space to service
            it or an object which has not yet been mapped. If all objects have been mapped and are full, a new extern
            object is mapped with a fixed size.
        """
        if self._extern_object is None:
            if self.main_object.arch.bits < 32:
                extern_size = 0x200
            elif self.main_object.arch.bits == 32:
                extern_size = 0x8000
            else:
                extern_size = 0x80000
            self._extern_object = ExternObject(self, map_size=extern_size)
            self._internal_load(self._extern_object)
        return self._extern_object

    @property
    def kernel_object(self) -> KernelObject:
        """
        Return the object used to provide addresses to syscalls.

        Accessing this property will load this object into memory if it was not previously present.
        """
        if self._kernel_object is None:
            self._kernel_object = KernelObject(self)
            self._map_object(self._kernel_object)
        return self._kernel_object

    @property
    def all_elf_objects(self) -> list[MetaELF]:
        """
        Return a list of every object that was loaded from an ELF file.
        """
        return [o for o in self.all_objects if isinstance(o, MetaELF)]

    @property
    def all_pe_objects(self) -> list[PE]:
        """
        Return a list of every object that was loaded from an ELF file.
        """
        return [o for o in self.all_objects if isinstance(o, PE)]

    @property
    def missing_dependencies(self) -> set[str]:
        """
        Return a set of every name that was requested as a shared object dependency but could not be loaded
        """
        return self.requested_names - {k for k, v in self._satisfied_deps.items() if v is not False}

    @property
    def auto_load_libs(self) -> bool:
        return self._auto_load_libs

    def describe_addr(self, addr: int) -> str:
        """
        Returns a textual description of what's in memory at the provided address
        """
        o = self.find_object_containing(addr)

        if o is None:
            return "not part of a loaded object"

        options = []

        rva = AT.from_va(addr, o).to_rva()

        idx = o.symbols.bisect_key_right(rva) - 1  # type: ignore
        while idx >= 0:
            sym = o.symbols[idx]
            if not sym.name or sym.is_import:
                idx -= 1
                continue
            options.append((sym.relative_addr, f"{sym.name}+"))
            break

        if isinstance(o, ELF):
            try:
                plt_addr, plt_name = max((a, n) for n, a in o._plt.items() if a <= rva)
            except ValueError:
                pass
            else:
                options.append((plt_addr, f"PLT.{plt_name}+"))

        options.append((0, "offset "))

        if o.provides:
            objname = o.provides
        elif o.binary:
            objname = os.path.basename(o.binary)
        elif self.main_object is o:
            objname = "main binary"
        else:
            objname = "object loaded from stream"

        best_offset, best_prefix = max(options, key=lambda v: v[0])
        return f"{best_prefix}{rva - best_offset:#x} in {objname} ({AT.from_va(addr, o).to_lva():#x})"

    # Search functions

    def find_object(self, spec: Backend | str, extra_objects: Iterable[Backend] = ()) -> Backend | None:
        """
        If the given library specification has been loaded, return its object, otherwise return None.
        """
        if isinstance(spec, Backend):
            for obj in self.all_objects:
                if obj is spec:
                    return obj
            return None

        if self._case_insensitive:
            spec = spec.lower()
        extra_idents = {}
        for obj in extra_objects:
            for ident in self._possible_idents(obj):
                extra_idents[ident] = obj

        for ident in self._possible_idents(spec):
            if ident in self._satisfied_deps:
                result = self._satisfied_deps[ident]
                if result is not False:
                    return result
            if ident in extra_idents:
                return extra_idents[ident]

        return None

    def find_object_containing(self, addr: int, membership_check: bool = True) -> Backend | None:
        """
        Return the object that contains the given address, or None if the address is unmapped.

        :param int addr:    The address that should be contained in the object.
        :param bool membership_check:   Whether a membership check should be performed or not (True by default). This
                                        option can be set to False if you are certain that the target object does not
                                        have "holes".
        :return:            The object or None.
        """

        def _check_object_memory(obj_):
            if isinstance(obj_.memory, Clemory):
                if AT.from_va(addr, obj_).to_rva() in obj_.memory:
                    self._last_object = obj_
                    return obj_
                return None
            elif isinstance(obj_.memory, str):
                self._last_object = obj_
                return obj_
            else:
                raise CLEError(f"Unsupported memory type {type(obj_.memory)}")

        # check the cache first
        if self._last_object is not None and self._last_object.min_addr <= addr <= self._last_object.max_addr:
            if not membership_check:
                return self._last_object
            if not self._last_object.has_memory:
                return self._last_object
            o = _check_object_memory(self._last_object)
            if o:
                return o

        if addr > self.max_addr or addr < self.min_addr:
            return None

        obj = key_bisect_floor_key(self.all_objects, addr, keyfunc=lambda x: x.min_addr)
        if obj is None:
            return None
        if not obj.min_addr <= addr <= obj.max_addr:
            return None
        if not membership_check:
            self._last_object = obj
            return obj
        if not obj.has_memory:
            self._last_object = obj
            return obj
        return _check_object_memory(obj)

    def find_segment_containing(self, addr: int, skip_pseudo_objects: bool = True) -> Segment | None:
        """
        Find the section object that the address belongs to.

        :param int addr: The address to test
        :param bool skip_pseudo_objects: Skip objects that CLE adds during loading.
        :return: The section that the address belongs to, or None if the address does not belong to any section, or if
                section information is not available.
        :rtype: cle.Segment
        """

        obj = self.find_object_containing(addr, membership_check=False)

        if obj is None:
            return None

        if skip_pseudo_objects and isinstance(obj, ExternObject | KernelObject | TLSObject):
            # the address is from a section allocated by angr.
            return None

        return obj.find_segment_containing(addr)

    def find_section_containing(self, addr: int, skip_pseudo_objects=True) -> Section | None:
        """
        Find the section object that the address belongs to.

        :param int addr: The address to test.
        :param bool skip_pseudo_objects: Skip objects that CLE adds during loading.
        :return: The section that the address belongs to, or None if the address does not belong to any section, or if
                section information is not available.
        :rtype: cle.Section
        """

        obj = self.find_object_containing(addr, membership_check=False)

        if obj is None:
            return None

        if skip_pseudo_objects and isinstance(obj, ExternObject | KernelObject | TLSObject):
            # the address is from a special CLE section
            return None

        return obj.find_section_containing(addr)

    def find_loadable_containing(self, addr: int, skip_pseudo_objects=True) -> Region | None:
        """
        Find the section or segment object the address belongs to. Sections will only be used if the corresponding
        object does not have segments.

        :param addr: The address to test
        :param skip_pseudo_objects: Skip objects that CLE adds during loading.
        :return:  The section or segment that the address belongs to, or None if the address does not belong to any
                    section or segment.
        """
        obj = self.find_object_containing(addr, membership_check=False)

        if obj is None:
            return None

        if skip_pseudo_objects and isinstance(obj, ExternObject | KernelObject | TLSObject):
            # the address is from a special CLE section
            return None

        return obj.find_loadable_containing(addr)

    def find_section_next_to(self, addr: int, skip_pseudo_objects=True) -> Section | None:
        """
        Find the next section after the given address.

        :param int addr: The address to test.
        :param bool skip_pseudo_objects: Skip objects that CLE adds during loading.
        :return: The next section that goes after the given address, or None if there is no section after the address,
                 or if section information is not available.
        :rtype: cle.Section
        """

        obj = self.find_object_containing(addr, membership_check=False)

        if obj is None:
            return None

        if skip_pseudo_objects and isinstance(obj, ExternObject | KernelObject | TLSObject):
            # the address is from a special CLE section
            return None

        return obj.sections.find_region_next_to(addr)

    def find_symbol(self, thing, fuzzy=False) -> Symbol | None:
        """
        Search for the symbol with the given name or address.

        :param thing:       Either the name or address of a symbol to look up
        :param fuzzy:       Set to True to return the first symbol before or at the given address

        :returns:           A :class:`cle.backends.Symbol` object if found, None otherwise.
        """
        if isinstance(thing, archinfo.arch_soot.SootAddressDescriptor):
            # Soot address
            # TODO launch this shit into the sun
            return thing.method.fullname  # type: ignore
        elif isinstance(thing, int):
            # address
            if fuzzy:
                so = self.find_object_containing(thing)
                if so is None:
                    return None
                objs = [so]
            else:
                objs = self.all_objects

            for so in objs:
                idx = so.symbols.bisect_key_right(AT.from_mva(thing, so).to_rva()) - 1  # type: ignore
                while idx >= 0 and (fuzzy or so.symbols[idx].rebased_addr == thing):
                    if so.symbols[idx].is_import:
                        idx -= 1
                        continue
                    return so.symbols[idx]
        else:
            # name
            for so in self.all_objects:
                if so is self._extern_object:
                    continue
                sym = so.get_symbol(thing)
                if sym is None:
                    continue

                if sym.is_import:
                    if sym.resolvedby is not None:
                        if sym.resolvedby.is_forward and sym.resolvedby.resolvedby is not None:
                            return sym.resolvedby.resolvedby
                        return sym.resolvedby
                else:
                    if sym.is_forward and sym.resolvedby is not None:
                        return sym.resolvedby
                    return sym

            if self._extern_object is not None:
                sym = self.extern_object.get_symbol(thing)
                if sym is not None:
                    return sym

        return None

    @property
    def symbols(self) -> Iterator[Symbol]:
        peeks = []
        for so in self.all_objects:
            if so.symbols:
                i = iter(so.symbols)
                n = next(i)
                peeks.append((n, i))
        while peeks:
            element = min(
                peeks, key=lambda x: x[0].rebased_addr
            )  # if we don't do this it might crash on comparing iterators
            n, i = element
            idx = peeks.index(element)
            yield n
            try:
                peeks[idx] = next(i), i
            except StopIteration:
                peeks.pop(idx)

    def find_all_symbols(
        self, name: str, exclude_imports=True, exclude_externs=False, exclude_forwards=True
    ) -> Iterable[Symbol]:
        """
        Iterate over all symbols present in the set of loaded binaries that have the given name

        :param name:                The name to search for
        :param exclude_imports:     Whether to exclude import symbols. Default True.
        :param exclude_externs:     Whether to exclude symbols in the extern object. Default False.
        :param exclude_forwards:    Whether to exclude forward symbols. Default True.
        """
        for so in self.all_objects:
            for sym in so.symbols:
                if sym is None:
                    continue
                if sym.name != name:
                    continue
                if sym.is_import and exclude_imports:
                    continue
                if sym.owner is self._extern_object and exclude_externs:
                    continue
                if sym.is_forward and exclude_forwards:
                    continue

                yield sym

    def find_plt_stub_name(self, addr: int) -> str | None:
        """
        Return the name of the PLT stub starting at ``addr``.
        """
        so = self.find_object_containing(addr)
        if so is not None and isinstance(so, MetaELF):
            return so.reverse_plt.get(addr, None)
        return None

    def find_relevant_relocations(self, name: str) -> Iterator[Relocation]:
        """
        Iterate through all the relocations referring to the symbol with the given ``name``
        """
        for so in self.all_objects:
            for reloc in so.relocs:
                if reloc.symbol is not None:
                    if reloc.symbol.name == name:
                        yield reloc

    # Complicated stuff

    def perform_irelative_relocs(self, resolver_func):
        """
        Use this method to satisfy ``IRelative`` relocations in the binary that require execution of loaded code.

        Note that this does NOT handle ``IFunc`` symbols, which must be handled separately. (this could be changed, but
        at the moment it's desirable to support lazy IFunc resolution, since emulation is usually slow)

        :param resolver_func:   A callback function that takes an address, runs the code at that address, and returns
                                the return value from the emulated function.
        """
        for obj in self.all_objects:
            for resolver, dest in obj.irelatives:
                val = resolver_func(resolver)
                if val is not None:
                    obj.memory.pack_word(dest, val)

    def dynamic_load(self, spec):
        """
        Load a file into the address space. Note that the sematics of ``auto_load_libs`` and ``except_missing_libs``
        apply at all times.

        :param spec:    The path to the file to load. May be an absolute path, a relative path, or a name to search in
                        the load path.

        :return:        A list of all the objects successfully loaded, which may be empty if this object was previously
                        loaded. If the object specified in ``spec`` failed to load for any reason, including the file
                        not being found, return None.
        """
        try:
            return self._internal_load(spec)
        except CLEFileNotFoundError as e:
            log.warning("Dynamic load failed: %r", e)
            return None

    def get_loader_symbolic_constraints(self):
        """
        Do not use this method.
        """
        if not self.aslr:
            return []

        try:
            import claripy  # pylint:disable=import-outside-toplevel
        except ImportError:
            claripy = None

        if not claripy:
            log.error("Please install claripy to get symbolic constraints")
            return []
        outputlist = []
        for obj in self.all_objects:
            # TODO Fix Symbolic for tls whatever
            if obj.aslr and isinstance(obj.mapped_base_symbolic, claripy.ast.BV):
                outputlist.append(obj.mapped_base_symbolic == obj.mapped_base)
        return outputlist

    # Private stuff

    @staticmethod
    def _is_linux_loader_name(name):
        """
        ld can have different names such as ld-2.19.so or ld-linux-x86-64.so.2 depending on symlinks and whatnot.
        This determines if `name` is a suitable candidate for ld.
        """
        return "ld.so" in name or "ld64.so" in name or "ld-linux" in name

    def _internal_load(self, *args, preloading=()):
        """
        Pass this any number of files or libraries to load. If it can't load any of them for any reason, it will
        except out. Note that the semantics of ``auto_load_libs`` and ``except_missing_libs`` apply at all times.

        It will return a list of all the objects successfully loaded, which may be smaller than the list you provided
        if any of them were previously loaded.

        The ``main_binary`` has to come first, followed by any additional libraries to load this round. To create the
        effect of "preloading", i.e. ensuring symbols are resolved to preloaded libraries ahead of any others, pass
        ``preloading`` as a list of identifiers which should be considered preloaded. Note that the identifiers will
        be compared using object identity.
        """
        # ideal loading pipeline:
        # - load everything, independently and recursively until dependencies are satisfied
        # - resolve symbol-based dependencies
        # - layout address space, including (as a prerequisite) coming up with the layout for tls and externs
        # - map everything into memory
        # - perform relocations

        # STEP 1
        # Load everything. for each binary, load it in isolation so we end up with a Backend instance.
        # If auto_load_libs is on, do this iteratively until all dependencies is satisfied
        objects = []
        preload_objects = []
        dependencies = []
        cached_failures = (
            set()
        )  # this assumes that the load path is global and immutable by the time we enter this func

        for main_spec in args:
            is_preloading = any(spec is main_spec for spec in preloading)
            if self.find_object(main_spec, extra_objects=objects) is not None:
                log.info("Skipping load request %s - already loaded", main_spec)
                continue
            obj = self._load_object_isolated(main_spec)
            objects.append(obj)
            objects.extend(obj.child_objects)
            dependencies.extend(obj.deps)

            if self._main_object is None:
                # this is technically the first place we can start to initialize things based on platform
                if obj.force_main_object is not None:
                    self._original_main_object = obj
                    self._main_object = obj.force_main_object
                    self._main_object.is_main_bin = True
                    # just to be safe, we clear obj.force_main_object here
                    obj.force_main_object = None
                else:
                    self._original_main_object = self._main_object = obj
                self._memory = Clemory(self._main_object.arch, root=True)

                chk_obj = (
                    self.main_object
                    if isinstance(self.main_object, ELFCore) or not self.main_object.child_objects
                    else self.main_object.child_objects[0]
                )
                if isinstance(chk_obj, ELFCore):
                    self._tls = ELFCoreThreadManager(self, self._main_object.arch)
                elif isinstance(self._main_object, Minidump):
                    self._tls = MinidumpThreadManager(self, self._main_object.arch)
                elif isinstance(chk_obj, MetaELF):
                    self._tls = ELFThreadManager(self, self._main_object.arch)
                elif isinstance(chk_obj, PE | Coff):
                    self._tls = PEThreadManager(self, self._main_object.arch)
                else:
                    self._tls = ThreadManager(self, self._main_object.arch)

            elif is_preloading:
                self.preload_libs.append(obj)
                preload_objects.append(obj)

        while self._auto_load_libs and dependencies:
            spec = dependencies.pop(0)
            if spec in cached_failures:
                log.debug("Skipping implicit dependency %s - cached failure", spec)
                continue
            if self.find_object(spec, extra_objects=objects) is not None:
                log.debug("Skipping implicit dependency %s - already loaded", spec)
                continue

            try:
                log.info("Loading %s...", spec)
                obj = self._load_object_isolated(spec)  # loading dependencies
            except CLEFileNotFoundError:
                log.info("... not found")
                cached_failures.add(spec)
                if self._except_missing_libs:
                    raise
                continue

            objects.append(obj)
            objects.extend(obj.child_objects)
            dependencies.extend(obj.deps)

            if isinstance(self.tls, ThreadManager):  # ... java
                if isinstance(obj, MetaELF):
                    self._tls = ELFThreadManager(self, obj.arch)
                elif isinstance(obj, PE):
                    self._tls = PEThreadManager(self, obj.arch)

        # STEP 1.5
        # produce dependency-ordered list of objects and soname map

        ordered_objects = []
        soname_mapping = OrderedDict(
            (obj.provides if not self._ignore_import_version_numbers else obj.provides.rstrip(".0123456789"), obj)
            for obj in objects
            if obj.provides
        )
        seen = set()

        def visit(obj):
            if id(obj) in seen:
                return
            seen.add(id(obj))

            stripped_deps = [
                dep if not self._ignore_import_version_numbers else dep.rstrip(".0123456789") for dep in obj.deps
            ]
            dep_objs = [soname_mapping[dep_name] for dep_name in stripped_deps if dep_name in soname_mapping]
            for dep_obj in dep_objs:
                visit(dep_obj)

            ordered_objects.append(obj)

        for obj in preload_objects + objects:
            visit(obj)

        # STEP 2
        # Resolve symbol dependencies. Create an unmapped extern object, which may not be used
        # after this step, everything should have the appropriate references to each other and the extern
        # object should have all the space it needs allocated

        extern_obj = ExternObject(self)

        # tls registration
        for obj in objects:
            self.tls.register_object(obj)

        # link everything
        for obj in ordered_objects:
            log.info("Linking %s", obj.binary)
            sibling_objs = list(obj.parent_object.child_objects) if obj.parent_object is not None else []
            stripped_deps = [
                dep if not self._ignore_import_version_numbers else dep.rstrip(".0123456789") for dep in obj.deps
            ]
            dep_objs = [soname_mapping[dep_name] for dep_name in stripped_deps if dep_name in soname_mapping]
            main_objs = [self.main_object] if self.main_object is not obj else []
            for reloc in obj.relocs:
                reloc.resolve_symbol(
                    main_objs + preload_objects + sibling_objs + dep_objs + [obj], extern_object=extern_obj
                )

        # if the extern object was used, add it to the list of objects we're mapping
        # also add it to the linked list of extern objects
        if extern_obj.map_size:
            # resolve the extern relocs this way because they may produce more relocations as we go
            i = 0
            while i < len(extern_obj.relocs):
                extern_obj.relocs[i].resolve_symbol(objects, extern_object=extern_obj)
                i += 1

            objects.append(extern_obj)
            ordered_objects.insert(0, extern_obj)
            extern_obj._next_object = self._extern_object
            self._extern_object = extern_obj

            extern_obj._finalize_tls()
            self.tls.register_object(extern_obj)

        # STEP 3
        # Map everything to memory
        for obj in objects:
            self._map_object(obj)

        # STEP 4
        # Perform relocations
        if self._perform_relocations:
            for obj in ordered_objects:
                obj.relocate()

        # Step 5
        # Insert each object into the appropriate mappings for lookup by name
        for obj in objects:
            self.requested_names.update(obj.deps)
            for ident in self._possible_idents(obj):
                self._satisfied_deps[ident] = obj

            if obj.provides is not None:
                self.shared_objects[obj.provides] = obj

        return objects

    def _load_object_isolated(self, spec, obj_ident: str | None = None):
        """
        Given a partial specification of a dependency, this will return the loaded object as a backend instance.
        It will not touch any loader-global data.
        """
        # STEP 1: identify file
        if isinstance(spec, Backend):
            return spec
        elif hasattr(spec, "read") and hasattr(spec, "seek"):
            binary_stream = spec
            binary = None
            close = False
        elif isinstance(spec, bytes | str):
            binary = self._search_load_path(spec)  # this is allowed to cheat and do partial static loading
            log.debug("... using full path %s", binary)
            binary_stream = open(binary, "rb")
            close = True
        elif isinstance(spec, Path):
            binary = str(spec)
            binary_stream = spec.open("rb")
            close = True
        else:
            raise CLEError(f"Bad library specification: {spec}")

        try:
            # STEP 2: collect options
            if self._main_object is None:
                options = dict(self._main_opts)
            else:
                for ident in itertools.chain(
                    [obj_ident], self._possible_idents(binary_stream if binary is None else binary)
                ):  # also allowed to cheat
                    if ident is None:
                        continue
                    if ident in self._lib_opts:
                        options = dict(self._lib_opts[ident])
                        break
                else:
                    options = {}

            # STEP 3: identify backend
            backend_spec = options.pop("backend", None)
            backend_cls = self._backend_resolver(backend_spec)
            if backend_cls is None:
                backend_cls = self._static_backend(binary_stream if binary is None else binary)
            if backend_cls is None:
                raise CLECompatibilityError(
                    f"Unable to find a loader backend for {spec}.  Perhaps try the 'blob' loader?"
                )

            # STEP 4: LOAD!
            log.debug("... loading with %s", backend_cls)

            result = backend_cls(binary, binary_stream, is_main_bin=self._main_object is None, loader=self, **options)
            result.close()
            return result
        finally:
            if close:
                binary_stream.close()

    def _map_object(self, obj: Backend):
        """
        This will integrate the object into the global address space, but will not perform relocations.
        """
        obj_size = obj.max_addr - obj.min_addr + 1

        if obj.pic:
            if obj._custom_base_addr is not None and self._is_range_free(obj._custom_base_addr, obj_size):
                base_addr = obj._custom_base_addr
            elif obj.linked_base and self._is_range_free(obj.linked_base, obj_size):
                base_addr = obj.linked_base
            elif not obj.is_main_bin:
                base_addr = self._find_safe_rebase_addr(obj_size)
            else:
                log.debug(
                    "The main binary is a position-independent executable. "
                    "It is being loaded with a base address of 0x400000."
                )
                base_addr = 0x400000

            obj.rebase(base_addr)
        else:
            if (
                obj._custom_base_addr is not None
                and obj.linked_base != obj._custom_base_addr
                and not isinstance(obj, Blob)
            ):
                log.warning(
                    "%s: base_addr was specified but the object is not PIC. " "specify force_rebase=True to override",
                    obj.binary_basename,
                )
            base_addr = obj.linked_base
            if not self._is_range_free(obj.linked_base, obj_size):
                raise CLEError(f"Position-DEPENDENT object {obj.binary} cannot be loaded at {base_addr:#x}")

        assert obj.mapped_base >= 0

        if obj.has_memory:
            assert obj.min_addr <= obj.max_addr
            log.info("Mapping %s at %#x", obj.binary, base_addr)
            self.memory.add_backer(base_addr, obj.memory)
        obj._is_mapped = True
        key_bisect_insort_right(self.all_objects, obj, keyfunc=lambda x: x.min_addr)

    # Address space management

    def _find_safe_rebase_addr(self, size):
        """
        Return a "safe" virtual address to map an object of size ``size``, i.e. one that won't
        overlap with anything already loaded.
        """
        # this assumes that self.main_object exists, which should... definitely be safe
        if self.main_object.arch.bits < 32 or self.main_object.max_addr >= 2 ** (self.main_object.arch.bits - 1):
            # HACK: On small arches, we should be more aggressive in packing stuff in.
            gap_start = 0
        else:
            gap_start = ALIGN_UP(self.main_object.max_addr + 1, self._rebase_granularity)
        for o in self.all_objects:
            if gap_start + size <= o.min_addr:
                break
            else:
                gap_start = ALIGN_UP(o.max_addr + 1, self._rebase_granularity)

        if gap_start + size > 2**self.main_object.arch.bits:
            # this may happen when loading an ELF core whose main object may occupy a large range of memory addresses
            # with large unoccupied holes left in the middle
            # we fall back to finding unoccupied holes
            for this_seg, next_seg in zip(self.main_object.segments.raw_list, self.main_object.segments.raw_list[1:]):
                gap_start = ALIGN_UP(this_seg.vaddr + this_seg.memsize, self._rebase_granularity)
                gap = next_seg.vaddr - gap_start
                if gap >= size:
                    break
            else:
                raise CLEOperationError("Ran out of room in address space")

        return gap_start

    def _is_range_free(self, va, size):
        # self.main_object should not be None here
        if va < 0 or va + size > 2**self.main_object.arch.bits:
            return False

        for o in self.all_objects:
            if o.min_addr <= va <= o.max_addr or va <= o.min_addr < va + size:
                return False

        return True

    # Functions of the form "use some heuristic to tell me about this spec"

    def _search_load_path(self, spec):
        """
        This will return the most likely full path that could satisfy the given partial specification.

        It will prefer files of a known filetype over files of an unknown filetype.
        """
        # this could be converted to being an iterator pretty easily
        for path in self._possible_paths(spec):
            if self._main_object is not None:
                backend_cls = self._static_backend(path)
                if backend_cls is None:
                    continue
                # If arch of main object is Soot ...
                if isinstance(self.main_object.arch, ArchSoot):
                    # ... skip compatibility check, since it always evaluates to false
                    # with native libraries (which are the only valid dependencies)
                    return path
                if not backend_cls.check_compatibility(path, self.main_object):
                    continue

            return path

        raise CLEFileNotFoundError(f"Could not find file {spec}")

    def _possible_paths(self, spec):
        """
        This iterates through each possible path that could possibly be used to satisfy the specification.

        The only check performed is whether the file exists or not.
        """
        dirs = []
        dirs.extend(self._custom_ld_path)  # if we say dirs = blah, we modify the original

        if self._main_object is not None:
            # add path of main binary
            if self.main_object.binary is not None:
                dirs.append(os.path.dirname(self.main_object.binary))
            # if arch of main_object is Soot ...
            is_arch_soot = isinstance(self.main_object.arch, ArchSoot)
            if is_arch_soot:
                # ... extend with load path of native libraries
                dirs.extend(self.main_object.extra_load_path)
                if self._use_system_libs:
                    log.debug(
                        "Path to system libraries (usually added as dependencies of JNI libs) needs "
                        "to be specified manually, by using the custom_ld_path option."
                    )
            # add path of system libraries
            if self._use_system_libs and not is_arch_soot:
                # Ideally this should be taken into account for each shared
                # object, not just the main object.
                dirs.extend(self.main_object.extra_load_path)
                if sys.platform.startswith("linux"):
                    dirs.extend(self.main_object.arch.library_search_path())
                elif sys.platform.startswith("openbsd"):
                    dirs.extend(self.main_object.arch.library_search_path())
                    dirs.extend(["/usr/local/lib", "/usr/X11R6/lib"])
                elif sys.platform == "win32":
                    native_dirs = os.environ["PATH"].split(";")

                    # simulate the wow64 filesystem redirect, working around the fact that WE may be impacted by it as
                    # a 32-bit python process.......
                    python_is_32bit = platform.architecture()[0] == "32bit"
                    guest_is_32bit = self.main_object.arch.bits == 32

                    if python_is_32bit != guest_is_32bit:
                        redirect_dir = os.path.join(os.environ["SystemRoot"], "system32").lower()
                        target_dir = os.path.join(
                            os.environ["SystemRoot"], "SysWOW64" if guest_is_32bit else "sysnative"
                        )
                        i = 0
                        while i < len(native_dirs):
                            if native_dirs[i].lower().startswith(redirect_dir):
                                # replace the access to System32 with SysWOW64 or sysnative
                                native_dirs[i] = target_dir + native_dirs[i][len(target_dir) :]
                            i += 1

                    dirs.extend(native_dirs)

        dirs.append(".")

        if self._case_insensitive:
            spec = spec.lower()

        for libdir in dirs:
            if self._case_insensitive:
                insensitive_path = self._path_insensitive(os.path.join(libdir, spec))
                if insensitive_path is not None:
                    yield os.path.realpath(insensitive_path)
            else:
                fullpath = os.path.realpath(os.path.join(libdir, spec))
                if os.path.exists(fullpath):
                    yield fullpath

            if self._ignore_import_version_numbers:
                try:
                    for libname in os.listdir(libdir):
                        ilibname = libname.lower() if self._case_insensitive else libname
                        if ilibname.strip(".0123456789") == spec.strip(".0123456789"):
                            yield os.path.realpath(os.path.join(libdir, libname))
                except OSError:
                    pass

    @classmethod
    def _path_insensitive(cls, path):
        """
        Get a case-insensitive path for use on a case sensitive system, or return None if it doesn't exist.

        From https://stackoverflow.com/a/8462613
        """
        if path == "" or os.path.exists(path):
            return path
        base = os.path.basename(path)  # may be a directory or a file
        dirname = os.path.dirname(path)
        suffix = ""
        if not base:  # dir ends with a slash?
            if len(dirname) < len(path):
                suffix = path[: len(path) - len(dirname)]
            base = os.path.basename(dirname)
            dirname = os.path.dirname(dirname)
        if not os.path.exists(dirname):
            dirname = cls._path_insensitive(dirname)
            if not dirname:
                return None
        # at this point, the directory exists but not the file
        try:  # we are expecting dirname to be a directory, but it could be a file
            files = os.listdir(dirname)
        except OSError:
            return None
        baselow = base.lower()
        try:
            basefinal = next(fl for fl in files if fl.lower() == baselow)
        except StopIteration:
            return None
        if basefinal:
            return os.path.join(dirname, basefinal) + suffix
        else:
            return None

    def _possible_idents(self, spec, lowercase=False):
        """
        This iterates over all the possible identifiers that could be used to describe the given specification.
        """
        if isinstance(spec, Backend):
            if spec.provides is not None:
                yield spec.provides
                if self._ignore_import_version_numbers:
                    yield spec.provides.rstrip(".0123456789")
            if spec.binary:
                yield spec.binary
                yield os.path.basename(spec.binary)
                yield os.path.basename(spec.binary).split(".")[0]
                if self._ignore_import_version_numbers:
                    yield os.path.basename(spec.binary).rstrip(".0123456789")
        elif hasattr(spec, "read") and hasattr(spec, "seek"):
            backend_cls = self._static_backend(spec, ignore_hints=True)
            if backend_cls is not None:
                soname = backend_cls.extract_soname(spec)
                if soname is not None:
                    yield soname
                    if self._ignore_import_version_numbers:
                        yield soname.rstrip(".0123456789")
        elif isinstance(spec, bytes | str):
            yield spec
            yield os.path.basename(spec)
            yield os.path.basename(spec).split(".")[0]
            if self._ignore_import_version_numbers:
                yield os.path.basename(spec).rstrip(".0123456789")

            if os.path.exists(spec):
                backend_cls = self._static_backend(spec, ignore_hints=True)
                if backend_cls is not None:
                    soname = backend_cls.extract_soname(spec)
                    if soname is not None:
                        yield soname
                        if self._ignore_import_version_numbers:
                            yield soname.rstrip(".0123456789")

        if not lowercase and (sys.platform == "win32" or self._case_insensitive):
            for name in self._possible_idents(spec, lowercase=True):
                yield name.lower()

    def _static_backend(self, spec, ignore_hints=False) -> type[Backend] | None:
        """
        Returns the correct loader for the file at `spec`.
        Returns None if it's a blob or some unknown type.
        TODO: Implement some binwalk-like thing to carve up blobs automatically
        """

        if not ignore_hints:
            for ident in self._possible_idents(spec):
                try:
                    return self._backend_resolver(self._lib_opts[ident]["backend"])
                except KeyError:
                    pass

        with stream_or_path(spec) as stream:
            for rear in [bk for bk in ALL_BACKENDS.values() if bk is not Blob] + [Blob]:
                if rear.is_default and rear.is_compatible(stream):
                    return rear

        return None

    @staticmethod
    def _backend_resolver(backend: str | type[Backend], default: T | None = None) -> type[Backend] | T | None:
        if isinstance(backend, type) and issubclass(backend, Backend):
            return backend
        elif backend in ALL_BACKENDS:
            return ALL_BACKENDS[backend]
        elif backend is None:
            return default
        else:
            raise CLEError(f"Invalid backend: {backend}")

    #
    # Memory data loading methods
    #

    def gen_ro_memview(self) -> None:
        """
        Generate a read-only view of the memory, and update self._memory_ro_view for faster data loading. Please call
        this method again for updating the read-only view, or discard_ro_memview() to discard any previously generated
        read-only views.
        """
        if self.memory is not None:
            self._memory_ro_view = ClemoryReadOnlyView(self.memory._arch, self.memory)

    def discard_ro_memview(self) -> None:
        """
        Discard any previously generated read-only views of the memory.
        """
        self._memory_ro_view = None

    def fast_memory_load_pointer(self, addr: int, size: int | None = None) -> int | None:
        """
        Perform a fast memory loading of a pointer.

        :param addr:    Address to read from.
        :param size:    Size of the pointer. Default to machine-word size.
        :return:        A pointer or None if the address does not exist.
        """

        try:
            mem = self.memory_ro_view if self.memory_ro_view is not None else self.memory
            return mem.unpack_word(addr, size=size)
        except KeyError:
            return None



from __future__ import annotations

from elftools.dwarf.die import DIE


class VariableType:
    """
    Entry class for DW_TAG_xxx_type

    :param name:            name of the type
    :param byte_size:       amount of bytes the type take in memory
    :param elf_object:      elf object to reference to (useful for pointer,...)

    :ivar name:             name of the type
    :type name:             str
    :ivar byte_size:        amount of bytes the type take in memory
    """

    def __init__(self, name: str, byte_size: int, elf_object):
        self.name = name
        self.byte_size = byte_size
        self._elf_object = elf_object

    @staticmethod
    def read_from_die(die: DIE, elf_object):
        """
        entry method to read a DW_TAG_xxx_type
        """
        if die.tag == "DW_TAG_base_type":
            return BaseType.read_from_die(die, elf_object)
        elif die.tag == "DW_TAG_pointer_type":
            return PointerType.read_from_die(die, elf_object)
        elif die.tag == "DW_TAG_structure_type":
            return StructType.read_from_die(die, elf_object)
        elif die.tag == "DW_TAG_array_type":
            return ArrayType.read_from_die(die, elf_object)
        elif die.tag == "DW_TAG_typedef":
            return TypedefType.read_from_die(die, elf_object)
        elif die.tag == "DW_TAG_union_type":
            return UnionType.read_from_die(die, elf_object)
        return None

    @staticmethod
    def supported_die(die: DIE) -> bool:
        return die.tag in (
            "DW_TAG_base_type",
            "DW_TAG_pointer_type",
            "DW_TAG_structure_type",
            "DW_TAG_array_type",
            "DW_TAG_typedef",
            "DW_TAG_union_type",
        )


class PointerType(VariableType):
    """
    Entry class for DW_TAG_pointer_type. It is inherited from VariableType

    :param byte_size:       amount of bytes the type take in memory
    :param elf_object:      elf object to reference to (useful for pointer,...)
    :param referenced_offset:  type of the referenced as offset in the compilation_unit
    """

    def __init__(self, byte_size: int, elf_object, referenced_offset: int):
        super().__init__("pointer", byte_size, elf_object)
        self._referenced_offset = referenced_offset

    @classmethod
    def read_from_die(cls, die: DIE, elf_object):
        """
        read an entry of DW_TAG_pointer_type. return None when there is no
        byte_size or type attribute.
        """
        byte_size = die.attributes.get("DW_AT_byte_size", None)

        if byte_size is None:
            return None

        dw_at_type = die.attributes.get("DW_AT_type", None)
        if dw_at_type is None:
            referenced_offset = None
        else:
            referenced_offset = dw_at_type.value + die.cu.cu_offset

        return cls(byte_size.value, elf_object, referenced_offset)

    @property
    def referenced_type(self):
        """
        attribute to get the referenced type. Return None if the type is not loaded
        """
        type_list = self._elf_object.type_list
        if self._referenced_offset in type_list.keys():
            return type_list[self._referenced_offset]
        return None


class BaseType(VariableType):
    """
    Entry class for DW_TAG_base_type. It is inherited from VariableType
    """

    # for __init__ see VariableType

    @classmethod
    def read_from_die(cls, die: DIE, elf_object):
        """
        read an entry of DW_TAG_base_type. return None when there is no
        byte_size attribute.
        """

        dw_at_name = die.attributes.get("DW_AT_name", None)
        byte_size = die.attributes.get("DW_AT_byte_size", None)
        if byte_size is None:
            return None
        return cls(dw_at_name.value.decode() if dw_at_name is not None else "unknown", byte_size.value, elf_object)


class StructType(VariableType):
    """
    Entry class for DW_TAG_structure_type. It is inherited from VariableType

    :param name:            name of the type
    :param byte_size:       amount of bytes the type take in memory
    :param elf_object:      elf object to reference to (useful for pointer,...)
    """

    def __init__(self, name: str, byte_size: int, elf_object, members):
        super().__init__(name, byte_size, elf_object)
        self.members = members

    @classmethod
    def read_from_die(cls, die: DIE, elf_object):
        """
        read an entry of DW_TAG_structure_type. return None when there is no
        byte_size attribute.
        """

        dw_at_name = die.attributes.get("DW_AT_name", None)
        byte_size = die.attributes.get("DW_AT_byte_size", None)

        if byte_size is None:
            return None

        members = []
        for die_child in die.iter_children():
            if die_child.tag == "DW_TAG_member":
                members.append(StructMember.read_from_die(die_child, elf_object))

        return cls(
            dw_at_name.value.decode() if dw_at_name is not None else "unknown", byte_size.value, elf_object, members
        )

    def __getitem__(self, member_name):
        for member in self.members:
            if member.name == member_name:
                return member
        raise KeyError


class UnionType(StructType):
    """
    Entry class for DW_TAG_union_type. Inherits from StructType to make it trivial.
    """


class StructMember:
    """
    Entry class for DW_TAG_member. This is not a type but a named member inside a struct.
    Use the property `type` to get its variable type.

    :param name:            name of the member
    :param addr_offset:     address offset of the member in the struct
    :param elf_object:      elf object to reference to (useful for pointer,...)
    :param type_offset:     type as offset in the compilation_unit

    :ivar name:             name of the member
    """

    def __init__(self, name: str, addr_offset: int, type_offset, elf_object):
        self.name = name
        self.addr_offset = addr_offset
        self._elf_object = elf_object
        self._type_offset = type_offset

    @classmethod
    def read_from_die(cls, die: DIE, elf_object):
        """
        read an entry of DW_TAG_member_type. return None when there is no
        type attribute.
        """

        dw_at_name = die.attributes.get("DW_AT_name", None)
        dw_at_type = die.attributes.get("DW_AT_type", None)
        dw_at_memloc = die.attributes.get("DW_AT_data_member_location", None)
        name = None if dw_at_name is None else dw_at_name.value.decode()
        ty = None if dw_at_type is None else dw_at_type.value + die.cu.cu_offset

        # From the DWARF5 manual, page 118:
        #    The member entry corresponding to a data member that is defined in a structure,
        #    union or class may have either a DW_AT_data_member_location attribute or a
        #    DW_AT_data_bit_offset attribute. If the beginning of the data member is the
        #    same as the beginning of the containing entity then neither attribute is required.
        # TODO bit_offset
        addr_offset = 0 if dw_at_memloc is None else dw_at_memloc.value

        return cls(name, addr_offset, ty, elf_object)

    @property
    def type(self):
        """
        attribute to get the type of the member. Return None if the type is not loaded
        """

        type_list = self._elf_object.type_list
        if self._type_offset in type_list.keys():
            return type_list[self._type_offset]
        return None


class ArrayType(VariableType):
    """
    Entry class for DW_TAG_array_type. It is inherited from VariableType

    :param byte_size:          amount of bytes the type take in memory
    :param elf_object:         elf object to reference to (useful for pointer,...)
    :param element_offset:     type of the array elements as offset in the compilation_unit
    """

    def __init__(self, byte_size, elf_object, element_offset):
        super().__init__("array", byte_size, elf_object)
        self._element_offset = element_offset

    @classmethod
    def read_from_die(cls, die: DIE, elf_object):
        """
        read an entry of DW_TAG_array_type. return None when there is no
        type attribute.
        """

        dw_byte_size = die.attributes.get("DW_AT_byte_size", None)

        dw_at_type = die.attributes.get("DW_AT_type", None)
        if dw_at_type is None:
            return None
        return cls(
            dw_byte_size.value if dw_byte_size is not None else None, elf_object, dw_at_type.value + die.cu.cu_offset
        )

    @property
    def element_type(self):
        type_list = self._elf_object.type_list
        if self._element_offset in type_list.keys():
            return type_list[self._element_offset]
        return None


class TypedefType(VariableType):
    """
    Entry class for DW_TAG_typedef. Inherits from VariableType.

    :param name:        name of the new type
    :param elf_object:  elf object to reference to (useful for pointer,...)
    :param type_offset: type as offset in the compilation_unit
    """

    def __init__(self, name: str, byte_size, elf_object, type_offset):
        super().__init__(name, byte_size, elf_object)
        self._type_offset = type_offset

    @classmethod
    def read_from_die(cls, die: DIE, elf_object):
        """
        read an entry of DW_TAG_member_type. return None when there is no
        type attribute.
        """

        dw_at_name = die.attributes.get("DW_AT_name", None)
        dw_at_type = die.attributes.get("DW_AT_type", None)
        dw_at_byte_size = die.attributes.get("DW_AT_byte_size", None)
        name = None if dw_at_name is None else dw_at_name.value.decode()
        type_offset = None if dw_at_type is None else dw_at_type.value + die.cu.cu_offset
        byte_size = None if dw_at_byte_size is None else dw_at_byte_size.value

        return cls(name, byte_size, elf_object, type_offset)

    @property
    def type(self):
        """
        attribute to get the type of the member. Return None if the type is not loaded
        """

        type_list = self._elf_object.type_list
        if self._type_offset in type_list.keys():
            return type_list[self._type_offset]
        return None



from __future__ import annotations

import archinfo

from cle.errors import CLEError

from .backend import Backend, register_backend
from .region import Section, Segment

try:
    from xbe import Xbe
except ImportError:
    Xbe = None


class XBESection(Section):
    def __init__(self, name, file_offset, file_size, virtual_addr, virtual_size, xbe_sec):
        super().__init__(name, file_offset, virtual_addr, virtual_size)
        self.filesize = file_size
        self._xbe_sec = xbe_sec

    @property
    def is_readable(self):
        """
        Whether this section has read permissions
        """
        return True

    @property
    def is_writable(self):
        """
        Whether this section has write permissions
        """
        return (self._xbe_sec.header.flags & self._xbe_sec.header.flags.WRITABLE) != 0

    @property
    def is_executable(self):
        """
        Whether this section has execute permissions
        """
        return (self._xbe_sec.header.flags & self._xbe_sec.header.flags.EXECUTABLE) != 0

    @property
    def only_contains_uninitialized_data(self):
        """
        We load every section in, they're all initialized
        """
        return False


class XBE(Backend):
    """
    The main loader class for statically loading XBE executables.
    """

    is_default = True

    def __init__(self, *args, **kwargs):
        if Xbe is None:
            raise CLEError("Run `pip install pyxbe==1.0.3` to support loading XBE files")
        super().__init__(*args, **kwargs)
        self.set_arch(archinfo.arch_from_id("x86"))

        self.os = "xbox"
        if self.binary is None:
            self._xbe = Xbe(data=self._binary_stream.read())
        else:
            self._xbe = Xbe.from_file(self.binary)
        self._entry = self._xbe.entry_addr
        self._image_vmem = bytearray(self._xbe.header.image_size)
        self._min_addr = self._xbe.header.base_addr
        self._max_addr = self._xbe.header.base_addr + self._xbe.header.image_size - 1

        # Add header
        self._add_xbe_section(
            0,
            self._xbe.header.image_header_size,
            self._xbe.header.base_addr,
            self._xbe.header.image_header_size,
            self._xbe.header_data,
        )

        # Add each section
        for _, sec in self._xbe.sections.items():
            self._add_xbe_section(
                sec.header.raw_addr,
                sec.header.raw_size,
                sec.header.virtual_addr,
                sec.header.virtual_size,
                sec.data,
                sec,
            )

        self.memory.add_backer(0, bytes(self._image_vmem))
        self.mapped_base = self.linked_base = self._xbe.header.base_addr

    def _add_xbe_section(self, file_offset, file_size, virtual_addr, virtual_size, backer, sec=None):
        # Copy in section contents
        start = virtual_addr - self._xbe.header.base_addr
        end = start + file_size
        self._image_vmem[start:end] = backer

        # Create a segment and a section
        seg = Segment(file_offset, virtual_addr, file_size, virtual_size)
        self.segments.append(seg)

        if sec is not None:
            sec = XBESection(sec.name, file_offset, file_size, virtual_addr, virtual_size, sec)
            self.sections.append(sec)

    def close(self):
        super().close()
        del self._xbe

    @staticmethod
    def is_compatible(stream):
        stream.seek(0)
        identstring = stream.read(4)
        stream.seek(0)
        return identstring.startswith(b"XBEH")

    @property
    def min_addr(self):
        return self._min_addr

    @property
    def max_addr(self):
        return self._max_addr

    @classmethod
    def check_compatibility(cls, spec, obj):  # pylint: disable=unused-argument
        assert False
        return True


register_backend("xbe", XBE)



from __future__ import annotations

from enum import IntEnum


class LoadCommands(IntEnum):
    """
    Enum for all the Load Commands used inside the MachO Binary
    """

    # https://opensource.apple.com/source/cctools/cctools-949.0.1/include/mach-o/loader.h.auto.html
    LC_REQ_DYLD = 0x80000000

    LC_SEGMENT = 0x1
    LC_SYMTAB = 0x2
    LC_SYMSEG = 0x3
    LC_THREAD = 0x4
    LC_UNIXTHREAD = 0x5
    LC_LOADFVMLIB = 0x6
    LC_IDFVMLIB = 0x7
    LC_IDENT = 0x8
    LC_FVMFILE = 0x9
    LC_PREPAGE = 0xA
    LC_DYSYMTAB = 0xB
    LC_LOAD_DYLIB = 0xC
    LC_ID_DYLIB = 0xD
    LC_LOAD_DYLINKER = 0xE
    LC_ID_DYLINKER = 0xF
    LC_PREBOUND_DYLIB = 0x10

    LC_ROUTINES = 0x11
    LC_SUB_FRAMEWORK = 0x12
    LC_SUB_UMBRELLA = 0x13
    LC_SUB_CLIENT = 0x14
    LC_SUB_LIBRARY = 0x15
    LC_TWOLEVEL_HINTS = 0x16
    LC_PREBIND_CKSUM = 0x17

    LC_LOAD_WEAK_DYLIB = 0x18 | LC_REQ_DYLD

    LC_SEGMENT_64 = 0x19
    LC_ROUTINES_64 = 0x1A
    LC_UUID = 0x1B
    LC_RPATH = 0x1C | LC_REQ_DYLD
    LC_CODE_SIGNATURE = 0x1D
    LC_SEGMENT_SPLIT_INFO = 0x1E
    LC_REEXPORT_DYLIB = 0x1F | LC_REQ_DYLD
    LC_LAZY_LOAD_DYLIB = 0x20
    LC_ENCRYPTION_INFO = 0x21
    LC_DYLD_INFO = 0x22
    LC_DYLD_INFO_ONLY = 0x22 | LC_REQ_DYLD
    LC_LOAD_UPWARD_DYLIB = 0x23 | LC_REQ_DYLD
    LC_VERSION_MIN_MACOSX = 0x24
    LC_VERSION_MIN_IPHONEOS = 0x25
    LC_FUNCTION_STARTS = 0x26
    LC_DYLD_ENVIRONMENT = 0x27

    LC_MAIN = 0x28 | LC_REQ_DYLD
    LC_DATA_IN_CODE = 0x29
    LC_SOURCE_VERSION = 0x2A
    LC_DYLIB_CODE_SIGN_DRS = 0x2B
    LC_ENCRYPTION_INFO_64 = 0x2C
    LC_LINKER_OPTION = 0x2D
    LC_LINKER_OPTIMIZATION_HINT = 0x2E
    LC_VERSION_MIN_TVOS = 0x2F
    LC_VERSION_MIN_WATCHOS = 0x30
    LC_NOTE = 0x31
    LC_BUILD_VERSION = 0x32
    LC_DYLD_EXPORTS_TRIE = 0x33 | LC_REQ_DYLD
    LC_DYLD_CHAINED_FIXUPS = 0x34 | LC_REQ_DYLD


class MachoFiletype(IntEnum):
    """
    from mach-o/loader.h

    Constants for the filetype field of the mach_header
    """

    MH_OBJECT = 1  # relocatable object file
    MH_EXECUTE = 2  # demand paged executable file
    MH_FVMLIB = 3  # fixed VM shared library file
    MH_CORE = 4  # core file
    MH_PRELOAD = 5  # preloaded executable file
    MH_DYLIB = 6  # dynamically bound shared library
    MH_DYLINKER = 7  # dynamic link editor
    MH_BUNDLE = 8  # dynamically bound bundle file
    MH_DYLIB_STUB = 9  # shared library stub for static
    MH_DSYM = 10  # companion file with only debug
    MH_KEXT_BUNDLE = 11  # x86_64 kexts
    MH_FILESET = 12  # set of mach-o's


class MH_flags(IntEnum):
    """
    from mach-o/loader.h
    Constants for the flags field of the mach_header
    """

    # the object file has no undefined references
    MH_NOUNDEFS = 0x1
    # the object file is the output of an incremental link against a base file and can't be link edited again
    MH_INCRLINK = 0x2
    # the object file is input for the dynamic linker and can't be staticly link edited again
    MH_DYLDLINK = 0x4
    # the object file's undefined references are bound by the dynamic linker when loaded.
    MH_BINDATLOAD = 0x8
    # the file has its dynamic undefined references prebound.
    MH_PREBOUND = 0x10
    # the file has its read-only and read-write segments split
    MH_SPLIT_SEGS = 0x20
    # the shared library init routine is to be run lazily via catching memory faults to its writeable segments(obsolete)
    MH_LAZY_INIT = 0x40
    # the image is using two-level name space bindings
    MH_TWOLEVEL = 0x80
    # the executable is forcing all images to use flat name space bindings
    MH_FORCE_FLAT = 0x100
    # this umbrella guarantees no multiple defintions of symbols in its sub-images
    # so the two-level namespace hints can always be used.
    MH_NOMULTIDEFS = 0x200
    # do not have dyld notify the prebinding agent about this executable
    MH_NOFIXPREBINDING = 0x400
    # the binary is not prebound but can have its prebinding redone. only used when MH_PREBOUND is not set.
    MH_PREBINDABLE = 0x800
    # indicates that this binary binds to all two-level namespace modules of its dependent libraries.
    # only used when MH_PREBINDABLE and MH_TWOLEVEL are both set.
    MH_ALLMODSBOUND = 0x1000
    # safe to divide up the sections into sub-sections via symbols for dead code stripping
    MH_SUBSECTIONS_VIA_SYMBOLS = 0x2000
    # the binary has been canonicalized via the unprebind operation
    MH_CANONICAL = 0x4000
    # the final linked image contains external weak symbols
    MH_WEAK_DEFINES = 0x8000
    # the final linked image uses weak symbols
    MH_BINDS_TO_WEAK = 0x10000
    # When this bit is set, all stacks in the task will be given stack execution privilege.
    # Only used in MH_EXECUTE filetypes.
    MH_ALLOW_STACK_EXECUTION = 0x20000
    # When this bit is set, the binary declares it is safe for use in processes with uid zero
    MH_ROOT_SAFE = 0x40000
    # When this bit is set, the binary declares it is safe for use in processes when issetugid() is true
    MH_SETUID_SAFE = 0x80000
    # When this bit is set on a dylib,
    # the static linker does not need to examine dependent dylibs to see if any are re-exported
    MH_NO_REEXPORTED_DYLIBS = 0x100000
    # When this bit is set, the OS will load the main executable at a random address. Only used in MH_EXECUTE filetypes.
    MH_PIE = 0x200000
    # Only for use on dylibs.
    # When linking against a dylib that has this bit set, the static linker will automatically not create a
    # LC_LOAD_DYLIB load command to the dylib if no symbols are being referenced from the dylib.
    MH_DEAD_STRIPPABLE_DYLIB = 0x400000
    # Contains a section of type S_THREAD_LOCAL_VARIABLES
    MH_HAS_TLV_DESCRIPTORS = 0x800000
    # When this bit is set, the OS will run the main executable with a non-executable heap
    # even on platforms (e.g. i386) that don't require it. Only used in MH_EXECUTE filetypes.
    MH_NO_HEAP_EXECUTION = 0x1000000
    # The code was linked for use in an application extension.
    MH_APP_EXTENSION_SAFE = 0x02000000
    # The nlist symbol table entries and the external relocation entries refer to the actual symbols and not
    # the dummy symbols.
    # This means that the external relocation entries don't need to be modified when symbols are added or removed.
    # This flag is only used in MH_OBJECT files.
    MH_NLIST_OUTOFSYNC_WITH_DYLDINFO = 0x04000000
    # The object file is a simulator wrapper file.
    MH_SIM_SUPPORT = 0x08000000
    # Only for use on dylibs.
    # When this bit is set, the dylib is part of the dyld shared cache, rather than loose in the filesystem
    MH_DYLIB_IN_CACHE = 0x80000000


class RebaseType(IntEnum):
    """
    from mach-o/loader.h
    """

    POINTER = 1
    TEXT_ABSOLUTE32 = 2
    TEXT_PCREL32 = 3


class RebaseOpcode(IntEnum):
    """
    from mach-o/loader.h

    #define REBASE_OPCODE_MASK					0xF0
    #define REBASE_IMMEDIATE_MASK					0x0F
    #define REBASE_OPCODE_DONE					0x00
    #define REBASE_OPCODE_SET_TYPE_IMM				0x10
    #define REBASE_OPCODE_SET_SEGMENT_AND_OFFSET_ULEB		0x20
    #define REBASE_OPCODE_ADD_ADDR_ULEB				0x30
    #define REBASE_OPCODE_ADD_ADDR_IMM_SCALED			0x40
    #define REBASE_OPCODE_DO_REBASE_IMM_TIMES			0x50
    #define REBASE_OPCODE_DO_REBASE_ULEB_TIMES			0x60
    #define REBASE_OPCODE_DO_REBASE_ADD_ADDR_ULEB			0x70
    #define REBASE_OPCODE_DO_REBASE_ULEB_TIMES_SKIPPING_ULEB	0x80
    """

    DONE = 0x00
    SET_TYPE_IMM = 0x10
    SET_SEGMENT_AND_OFFSET_ULEB = 0x20
    ADD_ADDR_ULEB = 0x30
    ADD_ADDR_IMM_SCALED = 0x40
    DO_REBASE_IMM_TIMES = 0x50
    DO_REBASE_ULEB_TIMES = 0x60
    DO_REBASE_ADD_ADDR_ULEB = 0x70
    DO_REBASE_ULEB_TIMES_SKIPPING_ULEB = 0x80

    @staticmethod
    def parse_byte(byte: int) -> tuple[RebaseOpcode, int]:
        """
        Split a byte into the RebaseOpcode and the immediate value
        :param byte:
        :return:
        """
        assert 0 <= byte <= 255
        return RebaseOpcode(byte & 0xF0), byte & 0x0F



# This file is part of Mach-O Loader for CLE.
# Contributed December 2016 by Fraunhofer SIT (https://www.sit.fraunhofer.de/en/) and updated in September 2019.
from __future__ import annotations

import logging
import struct
from collections.abc import Callable
from typing import TYPE_CHECKING

from cle.address_translator import AT
from cle.backends.relocation import Relocation
from cle.errors import CLEInvalidBinaryError

from .macho_enums import RebaseOpcode, RebaseType
from .symbol import AbstractMachOSymbol, BindingSymbol, DyldBoundSymbol, SymbolTableSymbol

if TYPE_CHECKING:
    from .macho import MachO

log = logging.getLogger(name=__name__)

OPCODE_MASK = 0xF0
IMM_MASK = 0x0F
BIND_TYPE_POINTER = 1
BIND_TYPE_TEXT_ABSOLUTE32 = 2
BIND_TYPE_TEXT_PCREL32 = 3
BIND_OPCODE_DONE = 0x00
BIND_OPCODE_SET_DYLIB_ORDINAL_IMM = 0x10
BIND_OPCODE_SET_DYLIB_ORDINAL_ULEB = 0x20
BIND_OPCODE_SET_DYLIB_SPECIAL_IMM = 0x30
BIND_OPCODE_SET_SYMBOL_TRAILING_FLAGS_IMM = 0x40
BIND_OPCODE_SET_TYPE_IMM = 0x50
BIND_OPCODE_SET_ADDEND_SLEB = 0x60
BIND_OPCODE_SET_SEGMENT_AND_OFFSET_ULEB = 0x70
BIND_OPCODE_ADD_ADDR_ULEB = 0x80
BIND_OPCODE_DO_BIND = 0x90
BIND_OPCODE_DO_BIND_ADD_ADDR_ULEB = 0xA0
BIND_OPCODE_DO_BIND_ADD_ADDR_IMM_SCALED = 0xB0
BIND_OPCODE_DO_BIND_ULEB_TIMES_SKIPPING_ULEB = 0xC0

if bytes is not str:

    def chh(x):
        return x

else:
    chh = ord


def read_uleb(blob: bytes, offset: int) -> tuple[int, int]:
    """Reads a number encoded as uleb128"""
    result = 0
    shift = 0
    index = offset

    while index < len(blob):
        b = chh(blob[index])
        result |= (b & 0x7F) << shift
        shift += 7
        index += 1
        if b & 0x80 == 0:
            break

    return result, index - offset


def read_sleb(blob, offset):
    """Reads a number encoded as sleb128"""
    result = 0
    shift = 0
    index = offset

    while index < len(blob):
        b = chh(blob[index])
        result |= (b & 0x7F) << shift
        shift += 7
        index += 1
        if b & 0x80 == 0:
            if b & 0x40:
                # two's complement
                result -= 1 << shift
            break

    return result, index - offset


class BindingState:
    """State object"""

    def __init__(self, is_64):
        self.index = 0
        self.done = False
        self.lib_ord = 0
        self.sym_name = ""
        self.sym_flags = 0
        self.binding_type = 0
        self.addend = 0
        self.segment_index = 0
        self.address = 0
        self.seg_end_address = 0  # TODO: no rebasing support
        # address is expected to properly overflow and address is uintptr_t (unsigned long according to _uintptr_t.h)
        self.wraparound = 2**64
        self.sizeof_intptr_t = 8 if is_64 else 4  # experimentally determined
        self.bind_handler = None  # function(state,binary) => None

    def add_address_ov(self, address, addend):
        """this is a very ugly klugde. It is needed because dyld relies on overflow
        semantics and represents several negative offsets through BIG ulebs"""
        tmp = address + addend
        if tmp > self.wraparound:
            tmp -= self.wraparound
        self.address = tmp

    def check_address_bounds(self):
        if self.address >= self.seg_end_address:
            log.error(
                "index %d: address >= seg_end_address (%#x >= %#x)", self.index, self.address, self.seg_end_address
            )
            raise CLEInvalidBinaryError()


class BindingHelper:
    """Factors out binding logic from MachO.
    Intended to work in close conjunction with MachO not for standalone use"""

    binary: MachO

    def __init__(self, binary):
        self.binary = binary

    def do_normal_bind(self, blob: bytes):
        """Performs non-lazy, non-weak bindings
        :param blob: Blob containing binding opcodes"""

        if blob is None:
            return  # skip

        log.debug("Binding non-lazy, non-weak symbols")
        s = BindingState(self.binary.arch.bits == 64)
        seg = self.binary.segments[0]
        s.seg_end_address = seg.vaddr + seg.memsize
        s.bind_handler = default_binding_handler
        self._do_bind_generic(
            blob,
            s,
            {
                0: n_opcode_done,
                0x10: n_opcode_set_dylib_ordinal_imm,
                0x20: n_opcode_set_dylib_ordinal_uleb,
                0x30: n_opcode_set_dylib_special_imm,
                0x40: n_opcode_set_trailing_flags_imm,
                0x50: n_opcode_set_type_imm,
                0x60: n_opcode_set_addend_sleb,
                0x70: n_opcode_set_segment_and_offset_uleb,
                0x80: n_opcode_add_addr_uleb,
                0x90: n_opcode_do_bind,
                0xA0: n_opcode_do_bind_add_addr_uleb,
                0xB0: n_opcode_do_bind_add_addr_imm_scaled,
                0xC0: n_opcode_do_bind_uleb_times_skipping_uleb,
            },
        )

        log.debug("Done binding non-lazy, non-weak symbols ")

    def do_lazy_bind(self, blob):
        """
        Performs lazy binding
        """
        if blob is None:
            return  # skip
        log.debug("Binding lazy symbols")

        s = BindingState(self.binary.arch.bits == 64)
        s.index = 0
        s.bind_handler = default_binding_handler
        end = len(blob)
        # We need to iterate the iteration as every lazy binding entry ends with BIND_OPCODE_DONE
        while s.index < end:
            # re-initialise state (except index)
            s.binding_type = 1
            s.address = 0
            s.sym_name = ""
            s.sym_flags = 0
            s.lib_ord = 0
            s.done = False
            s.addend = 0
            s.segment_index = 0
            s.seg_end_address = 0  # TODO: no rebasing support

            self._do_bind_generic(
                blob,
                s,
                {
                    0x00: n_opcode_done,
                    0x10: n_opcode_set_dylib_ordinal_imm,
                    0x20: n_opcode_set_dylib_ordinal_uleb,
                    0x30: n_opcode_set_dylib_special_imm,
                    0x40: n_opcode_set_trailing_flags_imm,
                    0x50: n_opcode_set_type_imm,
                    0x70: l_opcode_set_segment_and_offset_uleb,
                    0x90: l_opcode_do_bind,
                },
            )

        log.debug("Done binding lazy symbols")

    def do_rebases(self, blob: bytes):
        """
        Handles the rebase blob
        Implementation based closely on ImageLoaderMachOCompressed::rebase from dyld
        https://github.com/apple-opensource/dyld/blob/e3f88907bebb8421f50f0943595f6874de70ebe0/src/ImageLoaderMachOCompressed.cpp#L382-L463

        :param blob:
        :return:
        """
        if blob is None:
            return

        # State variables
        reloc_type: RebaseType | None = None
        done = False
        segment = None
        address = None
        index = 0
        end = len(blob)
        while not done and index < end:
            opcode, immediate = RebaseOpcode.parse_byte(blob[index])
            index += 1

            if opcode == RebaseOpcode.DONE:
                done = True

            elif opcode == RebaseOpcode.SET_TYPE_IMM:
                reloc_type = RebaseType(immediate)

            elif opcode == RebaseOpcode.SET_SEGMENT_AND_OFFSET_ULEB:
                segment = self.binary.segments[immediate]
                offset, index = self.read_uleb(blob, index)
                address = segment.vaddr + offset

            elif opcode == RebaseOpcode.ADD_ADDR_ULEB:
                uleb, index = self.read_uleb(blob, index)
                address += uleb

            elif opcode == RebaseOpcode.ADD_ADDR_IMM_SCALED:
                address += immediate * self.binary.arch.bytes

            elif opcode == RebaseOpcode.DO_REBASE_IMM_TIMES:
                for _ in range(immediate):
                    self.rebase_at(address, reloc_type)
                    address += self.binary.arch.bytes

            elif opcode == RebaseOpcode.DO_REBASE_ULEB_TIMES:
                count, index = self.read_uleb(blob, index)
                for _ in range(count):
                    if address >= segment.vaddr + segment.memsize:
                        raise CLEInvalidBinaryError()
                    self.rebase_at(address, reloc_type)
                    address += self.binary.arch.bytes

            elif opcode == RebaseOpcode.DO_REBASE_ADD_ADDR_ULEB:
                self.rebase_at(address, reloc_type)
                uleb, index = self.read_uleb(blob, index)
                address += uleb + self.binary.arch.bytes

            elif opcode == RebaseOpcode.DO_REBASE_ULEB_TIMES_SKIPPING_ULEB:
                count, index = self.read_uleb(blob, index)
                skip, index = self.read_uleb(blob, index)
                for _ in range(count):
                    if address >= segment.vaddr + segment.memsize:
                        raise CLEInvalidBinaryError()
                    self.rebase_at(address, reloc_type)
                    address += skip + self.binary.arch.bytes

            else:
                raise CLEInvalidBinaryError(f"Invalid opcode for current binding: {opcode:#x}")

    @staticmethod
    def read_uleb(blob, offset) -> tuple[int, int]:
        """
        little helper to read ulebs, that also returns the new index
        :param blob:
        :param offset:
        :return:
        """
        uleb, length = read_uleb(blob, offset)
        return uleb, offset + length

    def rebase_at(self, address: int, ty: RebaseType):
        relative_rebase_location = AT.from_lva(address, self.binary).to_rva()
        unslid_pointer = self.binary.memory.unpack_word(relative_rebase_location)
        relative_pointer = AT.from_lva(unslid_pointer, self.binary).to_rva()

        if ty == RebaseType.POINTER:
            reloc = MachOPointerRelocation(self.binary, relative_rebase_location, relative_pointer)
        elif ty == RebaseType.TEXT_ABSOLUTE32:
            reloc = MachOPointerRelocation(self.binary, relative_rebase_location, relative_pointer)
        elif ty == RebaseType.TEXT_PCREL32:
            raise NotImplementedError()
        else:
            raise ValueError(f"Invalid rebase type: {ty:#x}")
        self.binary.relocs.append(reloc)

    def _do_bind_generic(
        self,
        blob,
        init_state: BindingState,
        opcode_dict: dict[int, Callable[[BindingState, MachO, int, bytes], BindingState]],
    ):
        """
        Does the actual binding work. Represents a generic framework for interpreting binding opcodes
        :param blob: blob of binding opcodes
        :param init_state: Initial BindingState
        :param opcode_dict: Dictionary opcode=> handler
        :return: resulting binding state
        """
        s = init_state
        seg = self.binary.segments[s.segment_index]
        s.seg_end_address = seg.vaddr + seg.memsize  # TODO: no rebasing support
        end = len(blob)
        while not s.done and s.index < end:
            log.debug("Current address: %#x, blob index (offset): %#x", s.address, s.index)
            raw_opcode = blob[s.index]
            opcode = raw_opcode & OPCODE_MASK
            immediate = raw_opcode & IMM_MASK
            s.index += 1
            try:
                h = opcode_dict[opcode]
                s = h(s, self.binary, immediate, blob)
            except KeyError:
                log.error("Invalid opcode for current binding: %#x", opcode)

        return s


# pylint: disable=unused-argument
# The following functions realize different variants of handling binding opcodes
# the format is def X(state,binary,immediate,blob) => state
def n_opcode_done(s: BindingState, _b: MachO, _i: int, _blob: bytes) -> BindingState:
    log.debug("BIND_OPCODE_DONE @ %#x", s.index)
    s.done = True
    return s


def n_opcode_set_dylib_ordinal_imm(s: BindingState, _b: MachO, i: int, _blob: bytes) -> BindingState:
    log.debug("SET_DYLIB_ORDINAL_IMM @ %#x: %d", s.index, i)
    s.lib_ord = i
    return s


def n_opcode_set_dylib_ordinal_uleb(s: BindingState, _b: MachO, _i: int, blob: bytes) -> BindingState:
    uleb = read_uleb(blob, s.index)
    s.lib_ord = uleb[0]
    s.index += uleb[1]
    log.debug("SET_DYLIB_ORDINAL_ULEB @ %#x: %d", s.index, s.lib_ord)
    return s


def n_opcode_set_dylib_special_imm(s: BindingState, _b: MachO, i: int, _blob: bytes) -> BindingState:
    if i == 0:
        s.lib_ord = 0
    else:
        s.lib_ord = (i | OPCODE_MASK) - 256
    log.debug("SET_DYLIB_SPECIAL_IMM @ %#x: %d", s.index, s.lib_ord)
    return s


def n_opcode_set_trailing_flags_imm(s: BindingState, _b: MachO, i: int, blob: bytes) -> BindingState:
    s.sym_name = ""
    s.sym_flags = i

    while blob[s.index] != 0:
        s.sym_name += chr(blob[s.index])
        s.index += 1

    s.index += 1  # move past 0 byte
    log.debug("SET_SYMBOL_TRAILING_FLAGS_IMM @ %#x: %r,%#x", s.index - len(s.sym_name) - 1, s.sym_name, s.sym_flags)
    return s


def n_opcode_set_type_imm(s: BindingState, _b: MachO, i: int, _blob: bytes) -> BindingState:
    # pylint: disable=unused-argument
    s.binding_type = i
    log.debug("SET_TYPE_IMM @ %#x: %d", s.index, s.binding_type)
    return s


def n_opcode_set_addend_sleb(s: BindingState, _b: MachO, _i: int, blob: bytes) -> BindingState:
    sleb = read_sleb(blob, s.index)
    s.addend = sleb[0]
    log.debug("SET_ADDEND_SLEB @ %#x: %d", s.index, s.addend)
    s.index += sleb[1]
    return s


def n_opcode_set_segment_and_offset_uleb(s: BindingState, b: MachO, i: int, blob: bytes) -> BindingState:
    s.segment_index = i
    uleb = read_uleb(blob, s.index)
    log.debug("(n)SET_SEGMENT_AND_OFFSET_ULEB @ %#x: %d, %d", s.index, s.segment_index, uleb[0])
    s.index += uleb[1]
    seg = b.segments[s.segment_index]
    s.add_address_ov(seg.vaddr, uleb[0])
    s.seg_end_address = seg.vaddr + seg.memsize

    return s


def l_opcode_set_segment_and_offset_uleb(s: BindingState, b: MachO, i: int, blob: bytes) -> BindingState:
    uleb = read_uleb(blob, s.index)
    log.debug("(l)SET_SEGMENT_AND_OFFSET_ULEB @ %#x: %d, %d", s.index, i, uleb[0])
    seg = b.segments[i]
    s.add_address_ov(seg.vaddr, uleb[0])
    s.index += uleb[1]
    return s


def n_opcode_add_addr_uleb(s: BindingState, _b: MachO, _i: int, blob: bytes) -> BindingState:
    uleb = read_uleb(blob, s.index)
    s.add_address_ov(s.address, uleb[0])
    log.debug("ADD_ADDR_ULEB @ %#x: %d", s.index, uleb[0])
    s.index += uleb[1]
    return s


def n_opcode_do_bind(s: BindingState, b: MachO, _i: int, _blob: bytes) -> BindingState:
    log.debug("(n)DO_BIND @ %#x", s.index)
    s.check_address_bounds()
    s.bind_handler(s, b)
    s.add_address_ov(s.address, s.sizeof_intptr_t)
    return s


def l_opcode_do_bind(s: BindingState, b: MachO, _i: int, _blob: bytes) -> BindingState:
    log.debug("(l)DO_BIND @ %#x", s.index)
    s.bind_handler(s, b)
    return s


def n_opcode_do_bind_add_addr_uleb(s: BindingState, b: MachO, _i: int, blob: bytes) -> BindingState:
    uleb = read_uleb(blob, s.index)
    log.debug("DO_BIND_ADD_ADDR_ULEB @ %#x: %d", s.index, uleb[0])
    if s.address >= s.seg_end_address:
        log.error(
            "DO_BIND_ADD_ADDR_ULEB @ %#x: address >= seg_end_address (%#x>=%#x)", s.index, s.address, s.seg_end_address
        )
        raise CLEInvalidBinaryError()
    s.index += uleb[1]
    s.bind_handler(s, b)
    # this is done AFTER binding in preparation for the NEXT step
    s.add_address_ov(s.address, uleb[0] + s.sizeof_intptr_t)
    return s


def n_opcode_do_bind_add_addr_imm_scaled(s: BindingState, b: MachO, i: int, _blob: bytes) -> BindingState:
    log.debug("DO_BIND_ADD_ADDR_IMM_SCALED @ %#x: %d", s.index, i)
    if s.address >= s.seg_end_address:
        log.error(
            "DO_BIND_ADD_ADDR_IMM_SCALED @ %#x: address >= seg_end_address (%#x>=%#x)",
            s.index,
            s.address,
            s.seg_end_address,
        )
        raise CLEInvalidBinaryError()
    s.bind_handler(s, b)
    # this is done AFTER binding in preparation for the NEXT step
    s.add_address_ov(s.address, (i * s.sizeof_intptr_t) + s.sizeof_intptr_t)
    return s


def n_opcode_do_bind_uleb_times_skipping_uleb(s: BindingState, b: MachO, _i: int, blob: bytes) -> BindingState:
    count = read_uleb(blob, s.index)
    s.index += count[1]
    skip = read_uleb(blob, s.index)
    s.index += skip[1]
    log.debug("DO_BIND_ULEB_TIMES_SKIPPING_ULEB @ %#x: %d,%d", s.index - skip[1] - count[1], count[0], skip[0])
    for _ in range(0, count[0]):
        if s.address >= s.seg_end_address:
            log.error(
                "DO_BIND_ADD_ADDR_IMM_SCALED @ %#x: address >= seg_end_address (%#x >= %#x)",
                s.index - skip[1] - count[1],
                s.address,
                s.seg_end_address,
            )
            raise CLEInvalidBinaryError()
        s.bind_handler(s, b)
        s.add_address_ov(s.address, skip[0] + s.sizeof_intptr_t)
    return s


class MachOSymbolRelocation(Relocation):
    """
    Generic Relocation for MachO. It handles relocations that point to symbols
    """

    def __init__(self, owner: MachO, symbol: AbstractMachOSymbol, relative_addr: int, data):
        super().__init__(owner, symbol, relative_addr)
        self.data = data

    def resolve_symbol(self, solist, thumb=False, extern_object=None, **kwargs):
        if isinstance(self.symbol, SymbolTableSymbol | BindingSymbol | DyldBoundSymbol):
            for so in solist:
                if self.symbol.library_base_name == so.binary_basename:
                    symbols = so.get_symbol(self.symbol.name)
                    if not symbols:
                        log.warning(
                            "Symbol %s should have been resolved by %s but wasn't", self.symbol.name, so.binary_basename
                        )
                        continue
                    (symbol,) = symbols
                    assert symbol.is_export
                    self.resolve(symbol, extern_object=extern_object)
                    log.info("Resolved %s to %s", self.symbol.name, symbol)
                    return
            # None of the available libraries contain it, so we create an extern symbol for it
            new_symbol = extern_object.make_extern(self.symbol.name, sym_type=self.symbol._type, thumb=thumb)
            self.resolve(new_symbol, extern_object=extern_object)
        else:
            raise NotImplementedError("Did not expect this to happen")

    @property
    def dest_addr(self):
        return self.relative_addr

    @property
    def value(self):
        return self.resolvedby.rebased_addr

    def __repr__(self):
        return f"<MachO Reloc for {self.symbol} at {hex(self.relative_addr)}>"


class MachOPointerRelocation(Relocation):
    """
    A relocation for a pointer without any associated symbol
    These are either generated while handling the rebase blob, or while parsing chained fixups
    """

    def __init__(self, owner: MachO, relative_addr: int, data):
        """

        :param owner:
        :param relative_addr: the relative addr where this relocation is located
        :param data: the rebase offset relative to the linked base
        """
        super().__init__(owner, None, relative_addr)
        self.data = data

    @property
    def value(self):
        return self.owner.mapped_base + self.data

    def resolve_symbol(self, solist, thumb=False, extern_object=None, **kwargs):
        """
        This relocation has no associated symbol, so we don't need to resolve it.
        :param solist:
        :param thumb:
        :param extern_object:
        :param kwargs:
        :return:
        """
        # This needs to be set to true, so that the rebase will actually be applied later
        self.resolved = True

    def __repr__(self):
        return f"<MachO Ptr Fixup at {hex(self.relative_addr)} to {hex(self.data)}>"


# default binding handler
def default_binding_handler(state: BindingState, binary: MachO):
    """Binds location to the symbol with the given name and library ordinal"""

    # locate the symbol:
    matches = binary.symbols.get_by_name_and_ordinal(state.sym_name, state.lib_ord)
    if len(matches) > 1:
        log.error("Cannot bind: More than one match for (%r,%d)", state.sym_name, state.lib_ord)
        raise CLEInvalidBinaryError()
    if len(matches) < 1:
        log.info("No match for (%r,%d), generating BindingSymbol ...", state.sym_name, state.lib_ord)
        matches = [BindingSymbol(binary, state.sym_name, state.lib_ord)]
        binary.symbols.add(matches[0])
        binary._ordered_symbols.append(matches[0])

    symbol = matches[0]
    location = state.address

    # If the linked_addr is equal to zero, it's an imported symbol which is by that time unresolved.
    # Don't write addend's there

    value = symbol.linked_addr + state.addend if symbol.linked_addr != 0 else 0x0

    if state.binding_type == 1:  # POINTER
        log.debug("Updating address %#x with symobl %r @ %#x", location, state.sym_name, value)
        addr = AT.from_lva(location, binary).to_rva()
        data = struct.pack(binary.struct_byteorder + ("Q" if binary.arch.bits == 64 else "I"), value)
        reloc = MachOSymbolRelocation(binary, symbol, addr, data)
        binary.relocs.append(reloc)
        symbol.bind_xrefs.append(location)
    elif state.binding_type == 2:  # ABSOLUTE32
        location_32 = location % (2**32)
        value_32 = value % (2**32)
        log.debug("Updating address %#x with symobl %r @ %#x", state.sym_name, location_32, value_32)
        binary.memory.store(
            AT.from_lva(location_32, binary).to_rva(), struct.pack(binary.struct_byteorder + "I", value_32)
        )
        symbol.bind_xrefs.append(location_32)
    elif state.binding_type == 3:  # PCREL32
        location_32 = location % (2**32)
        value_32 = (value - (location + 4)) % (2**32)
        log.debug("Updating address %#x with symobl %r @ %#x", state.sym_name, location_32, value_32)
        binary.memory.store(
            AT.from_lva(location_32, binary).to_rva(), struct.pack(binary.struct_byteorder + "I", value_32)
        )
        symbol.bind_xrefs.append(location_32)
    else:
        log.error("Unknown BIND_TYPE: %d", state.binding_type)
        raise CLEInvalidBinaryError()



from __future__ import annotations

import logging

import archinfo

from .tls_object import ThreadManager

log = logging.getLogger(__name__)


class ELFCoreThreadManager(ThreadManager):
    def __init__(self, loader, arch, **kwargs):  # pylint: disable=unused-argument
        self.loader = loader
        self.arch = arch
        self.threads = [ELFCoreThread(loader, arch, threadinfo) for threadinfo in loader.main_object._threads]
        if arch.name not in ("AMD64", "X86"):
            log.warning("TLS for coredumps won't be right for this arch - idk how to do it")
        self.modules = []  # ???

    def new_thread(self, insert=False):  # pylint: disable=no-self-use
        raise TypeError("Cannot create new threads from a core file... for now")

    def register_object(self, obj):
        pass


class ELFCoreThread:
    def __init__(self, loader, arch: archinfo.Arch, threadinfo):
        self.loader = loader
        self.arch = arch
        self._threadinfo = threadinfo
        if arch.name == "AMD64":
            self.thread_pointer = threadinfo["registers"]["fs_base"]
        elif arch.name == "X86":
            gs = threadinfo["registers"]["gs"]
            if gs == 0:
                # I have no idea why this happens
                gs = next(iter(threadinfo["segments"].keys())) << 3
            self.thread_pointer = threadinfo["segments"][gs >> 3][0]
        else:
            self.thread_pointer = 0

        self.user_thread_pointer = self.thread_pointer + arch.elf_tls.tp_offset

    @property
    def dtv(self):
        return self.loader.memory.unpack_word(self.thread_pointer + self.arch.elf_tls.dtv_offsets[0])

    def get_addr(self, module_id, offset):
        """
        basically ``__tls_get_addr``.
        """
        return self.loader.memory.unpack_word(self.dtv + module_id * self.arch.bytes) + offset



from __future__ import annotations

from .variable import Variable


class LexicalBlock:
    """
    A lexical block is a sequence of source statements, e.g. a while/for
    loop or an if statement or some bracketed block.

    Corresponds to a DW_TAG_LexicalBlock in DWARF.

    :param super_block: The lexical block which contains this block
    :param low_pc:      The relative start address of the block
    :param high_pc:     The relative end address of the block

    :ivar low_pc:       The relative start address of the subprogram
    :ivar high_pc:      The relative end address of the subprogram
    :ivar child_blocks: Lexical blocks inside this block (only direct childs)
    :type child_blocks: List[LexicalBlock]
    """

    def __init__(self, low_pc, high_pc) -> None:
        self.low_pc = low_pc
        self.high_pc = high_pc
        self.child_blocks: list[LexicalBlock] = []


class Subprogram(LexicalBlock):
    """
    DW_TAG_subprogram for DWARF. The behavior is mostly inherited from
    LexicalBlock to avoid redundancy.

    :param name:     The name of the function/program
    :param low_pc:   The relative start address of the subprogram
    :param high_pc:  The relative end address of the subprogram

    :ivar name:            The name of the function/program
    :type name:            str
    :ivar local_variables: All local variables in a Subprogram (they may reside in serveral child blocks)
    :type local_variables: List[Variables]
    """

    def __init__(self, name, low_pc, high_pc) -> None:
        # pass self as the super_block of this subprogram
        self.subprogram = self
        super().__init__(low_pc, high_pc)
        self.name = name
        self.local_variables: list[Variable] = []
